"""
# Kernel sequence
"""

from dataclasses import dataclass
from typing import Sequence

import jax

from .epoch import EpochState
from .pytree import register_dataclass_as_pytree
from .types import (
    Kernel,
    KernelState,
    KeyArray,
    ModelState,
    Position,
    TransitionInfo,
    TuningInfo,
)

KernelStates = list[KernelState]
TuningInfos = dict[str, TuningInfo]
TransitionInfos = dict[str, TransitionInfo]


@register_dataclass_as_pytree
@dataclass
class KerSeqTransitionOutput:
    model_state: ModelState
    kernel_states: KernelStates
    infos: TransitionInfos


@register_dataclass_as_pytree
@dataclass
class KerSeqTuningOutput:
    kernel_states: KernelStates
    infos: TuningInfos


@register_dataclass_as_pytree
@dataclass
class KerSeqFinalizeWarmupOutput:
    kernel_states: KernelStates
    error_codes: dict[str, int]


class KernelSequence:
    def __init__(self, kernels: Sequence[Kernel]) -> None:
        identifiers = set()
        for ker in kernels:
            if not ker.identifier:
                raise RuntimeError(
                    f"Kernel identifier must be a non-empty string. "
                    f"The field is empty in {ker!r}."
                )
            identifiers.add(ker.identifier)

        if len(identifiers) < len(kernels):
            raise RuntimeError("Kernel identifier must be unique.")

        self._kernels = list(kernels)

    def get_kernels(self):
        return self._kernels

    def init_states(self, prng_key: KeyArray, model_state: ModelState) -> KernelStates:
        keys = jax.random.split(prng_key, len(self._kernels))
        states = [
            kernel.init_state(keys[i], model_state)
            for i, kernel in enumerate(self._kernels)
        ]
        return states

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_states: KernelStates,
        model_state: ModelState,
        epoch: EpochState,
    ) -> KernelStates:
        keys = jax.random.split(prng_key, len(self._kernels))

        states = [
            kernel.start_epoch(keys[i], kernel_states[i], model_state, epoch)
            for i, kernel in enumerate(self._kernels)
        ]
        return states

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_states: KernelStates,
        model_state: ModelState,
        epoch: EpochState,
    ) -> KernelStates:
        keys = jax.random.split(prng_key, len(self._kernels))

        states = [
            kernel.end_epoch(keys[i], kernel_states[i], model_state, epoch)
            for i, kernel in enumerate(self._kernels)
        ]
        return states

    def transition(
        self,
        prng_key: KeyArray,
        kernel_states: KernelStates,
        model_state: ModelState,
        epoch: EpochState,
    ) -> KerSeqTransitionOutput:
        keys = jax.random.split(prng_key, len(self._kernels))
        infos: TransitionInfos = {}
        kstates: KernelStates = []

        for i, kernel in enumerate(self._kernels):
            result = kernel.transition(keys[i], kernel_states[i], model_state, epoch)
            model_state = result.model_state
            kstates.append(result.kernel_state)
            infos[kernel.identifier] = result.info

        return KerSeqTransitionOutput(
            model_state=model_state, kernel_states=kstates, infos=infos
        )

    def tune(
        self,
        prng_key: KeyArray,
        kernel_states: KernelStates,
        model_state: ModelState,
        phase: EpochState,
        history: Position | None,
    ) -> KerSeqTuningOutput:
        keys = jax.random.split(prng_key, len(self._kernels))
        infos: TuningInfos = {}
        kstates = []

        # time = phase.time_before_epoch + phase.time_in_epoch

        for i, kernel in enumerate(self._kernels):
            result = kernel.tune(keys[i], kernel_states[i], model_state, phase, history)
            kstates.append(result.kernel_state)
            infos[kernel.identifier] = result.info

        return KerSeqTuningOutput(kernel_states=kstates, infos=infos)

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_states: KernelStates,
        model_state: ModelState,
        tuning_history: None | TuningInfos,
    ) -> KerSeqFinalizeWarmupOutput:
        keys = jax.random.split(prng_key, len(self._kernels))
        new_states = []
        error_codes: dict[str, int] = {}

        for i, kernel in enumerate(self._kernels):
            if tuning_history is None:
                th = None
            else:
                th = tuning_history[kernel.identifier]

            result = kernel.end_warmup(
                keys[i],
                kernel_states[i],
                model_state,
                th,
            )
            new_states.append(result.kernel_state)
            error_codes[kernel.identifier] = result.error_code

        return KerSeqFinalizeWarmupOutput(new_states, error_codes)



"""
# MCMC chains

This module is experimental. Expect API changes.
"""

from typing import Callable, Generic, Protocol, Sequence, TypeVar

from liesel.option import Option

from .epoch import EpochConfig
from .pytree import concatenate_leaves
from .types import PyTree

TPyTree = TypeVar("TPyTree", bound=PyTree)


class Chain(Protocol[TPyTree]):
    """
    A `Chain` stores multiple chucks of pytrees and concatenates them along a time
    axis.

    A `Chain` operates on the assumption that all chunks are pytrees with the same
    structure.

    The `Chain` can either represent one chain or multiple independent chains.
    For a single chain, the first dimension of each leaf represents the time.
    For a multi-chain, the first dimension is the chain index and the second dimension
    represents the time. Consequently, the leafs in the pytree must have a dimension of
    at least one (single chain) or two (multi-chain).

    For a multi-chain, the size of the first dimension (the chain index) is not
    allowed to vary among chunks.
    """

    @property
    def multichain(self) -> bool:
        """Returns true if the chain is a multi-chain, and false otherwise."""

    def append(self, chunk: TPyTree) -> None:
        """Appends a chunk to the chain."""

    def get(self) -> Option[TPyTree]:
        """
        Returns all chunks combined into one pytree.

        The option is none if no samples are in the chain.
        """


class EpochChain(Chain[TPyTree]):
    """An `EpochChain` is a `Chain` with an associated `EpochConfig`."""

    @property
    def epoch(self) -> EpochConfig:
        """Returns the associated `EpochConfig`."""


# implementations


class ListChain(Generic[TPyTree]):
    """Implements the `Chain` protocol with a list as storage."""

    def __init__(self, multichain: bool):
        self._multichain = multichain
        self._chunks_list: list[TPyTree] = []

    @property
    def multichain(self) -> bool:
        return self._multichain

    def append(self, chunk: TPyTree) -> None:
        return self._chunks_list.append(chunk)

    def _concatenate(self) -> None:
        time_axis = 1 if self.multichain else 0
        combined = concatenate_leaves(self._chunks_list, time_axis)
        if combined is not None:
            self._chunks_list = [combined]

    def get(self) -> Option[TPyTree]:
        if len(self._chunks_list) == 0:
            return Option(None)
        else:
            self._concatenate()
            return Option(self._chunks_list[0])


class ListEpochChain(ListChain[TPyTree]):
    """Implements the `EpochChain` protocol with a list as storage."""

    def __init__(self, multichain: bool, epoch: EpochConfig):
        super().__init__(multichain)
        self._epoch = epoch

    @property
    def epoch(self) -> EpochConfig:
        return self._epoch


class EpochChainManager(Generic[TPyTree]):
    """
    An `EpochChainManager` is a container for multiple epoch chains.

    The chains can be concatenated over multiple epochs.
    """

    def __init__(self, multichain: bool) -> None:
        self._multichain = multichain
        self._chains: list[ListEpochChain[TPyTree]] = []

    @property
    def current_epoch(self) -> EpochConfig:
        return self._chains[-1].epoch

    def advance_epoch(self, epoch: EpochConfig) -> None:
        new_chain: ListEpochChain[TPyTree] = ListEpochChain(self._multichain, epoch)
        self._chains.append(new_chain)

    def append(self, chunk: TPyTree) -> None:
        self._chains[-1].append(chunk)

    def get_epochs(self) -> Sequence[EpochConfig]:
        return [c.epoch for c in self._chains]

    def get_specific_chain(self, epoch_number: int) -> ListEpochChain[TPyTree]:
        return self._chains[epoch_number]

    def get_current_chain(self) -> ListEpochChain[TPyTree]:
        return self._chains[-1]

    def get_current_epoch(self) -> EpochConfig:
        return self._chains[-1].epoch

    def combine(self, epoch_numbers: Sequence[int]) -> Option[TPyTree]:
        chain: ListChain[TPyTree] = ListChain(self._multichain)
        for num in epoch_numbers:
            epoch_chain = self._chains[num]
            chunk = epoch_chain.get()
            if chunk.is_some():
                chain.append(chunk.unwrap())
        return chain.get()

    def combine_all(self) -> Option[TPyTree]:
        chain: ListChain[TPyTree] = ListChain(self._multichain)
        for epoch_chain in self._chains:
            chunk = epoch_chain.get()
            if chunk.is_some():
                chain.append(chunk.unwrap())
        return chain.get()

    def combine_filtered(
        self, predicate: Callable[[EpochConfig], bool]
    ) -> Option[TPyTree]:
        chain: ListChain[TPyTree] = ListChain(self._multichain)
        for epoch_chain in self._chains:
            if predicate(epoch_chain.epoch):
                chunk = epoch_chain.get()
                if chunk.is_some():
                    chain.append(chunk.unwrap())
        return chain.get()



"""
# Kernel-related info, outcome and mixin classes
"""

from abc import abstractmethod
from dataclasses import dataclass
from typing import Callable, Generic

import jax

from .epoch import EpochState, EpochType
from .pytree import register_dataclass_as_pytree
from .types import (
    KeyArray,
    ModelInterface,
    ModelState,
    Position,
    TKernelState,
    TTransitionInfo,
    TTuningInfo,
)


@register_dataclass_as_pytree
@dataclass
class DefaultTransitionInfo:
    error_code: int
    acceptance_prob: float
    position_moved: int

    def minimize(self):
        return DefaultTransitionInfo(
            self.error_code, self.acceptance_prob, self.position_moved
        )


@register_dataclass_as_pytree
@dataclass
class DefaultTuningInfo:
    error_code: int
    time: int


@register_dataclass_as_pytree
@dataclass
class TransitionOutcome(Generic[TKernelState, TTransitionInfo]):
    """
    A dataclass for the return value of the kernel method `transition`.
    Different kernels can use different types of `liesel.goose.types.KernelState`'s
    and `liesel.goose.types.TransitionInfo`'s.
    """

    info: TTransitionInfo
    kernel_state: TKernelState
    model_state: ModelState


@register_dataclass_as_pytree
@dataclass
class TuningOutcome(Generic[TKernelState, TTuningInfo]):
    """
    A dataclass for the return value of the kernel method `tune`.
    Different kernels can use different types of `liesel.goose.types.KernelState`'s
    and `liesel.goose.types.TuningInfo`'s.
    """

    info: TTuningInfo
    kernel_state: TKernelState


@register_dataclass_as_pytree
@dataclass
class WarmupOutcome(Generic[TKernelState]):
    """
    A dataclass for the return value of the kernel method `end_warmup`.
    Different kernels can use different types of `liesel.goose.types.KernelState`'s.
    """

    error_code: int
    kernel_state: TKernelState


class ModelMixin:
    """
    A mixin facilitating the interaction with the model interface.
    """

    position_keys: tuple[str, ...]
    _model: ModelInterface | None

    @property
    def model(self) -> ModelInterface:
        """Returns the model interface if it is set. Raises error otherwise."""

        if self._model is None:
            raise RuntimeError("Model interface not set")

        return self._model

    def set_model(self, model: ModelInterface):
        self._model = model

    def has_model(self) -> bool:
        return self._model is not None

    def position(self, model_state: ModelState) -> Position:
        """
        Extracts the position from a model state.
        """

        return self.model.extract_position(self.position_keys, model_state)

    def log_prob_fn(self, model_state: ModelState) -> Callable[[Position], float]:
        """
        Returns the log-probability function with the position as the only argument.
        """

        def log_prob_fn(position: Position) -> float:
            new_model_state = self.model.update_state(position, model_state)
            return self.model.log_prob(new_model_state)

        return log_prob_fn


class TransitionMixin(Generic[TKernelState, TTransitionInfo]):
    """
    An abstract mixin defining two transition methods with and without adaptation.
    """

    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        is_adaptation = EpochType.is_adaptation(epoch.config.type)

        outcome: TransitionOutcome[TKernelState, TTransitionInfo] = jax.lax.cond(
            is_adaptation,
            self._adaptive_transition,
            self._standard_transition,
            prng_key,
            kernel_state,
            model_state,
            epoch,
        )

        return outcome

    @abstractmethod
    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """
        Performs an MCMC transition *outside* an adaptation epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """
        Performs an MCMC transition *in* an adaptation epoch. Must be jittable.
        """

        raise NotImplementedError


class TuningMixin(Generic[TKernelState, TTuningInfo]):
    """
    An abstract mixin defining two tuning methods after a slow and a fast
    adaptation epoch.
    """

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        is_slow = epoch.config.type == EpochType.SLOW_ADAPTATION

        outcome: TuningOutcome[TKernelState, TTuningInfo] = jax.lax.cond(
            is_slow,
            self._tune_slow,
            self._tune_fast,
            prng_key,
            kernel_state,
            model_state,
            epoch,
            history,
        )

        return outcome

    @abstractmethod
    def _tune_fast(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        Tunes a kernel after a *fast* adaptation epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def _tune_slow(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        Tunes a kernel after a *slow* adaptation epoch. Must be jittable.
        """

        raise NotImplementedError



"""
# MCMC engine

This module is experimental. Expect API changes.
"""

# mypy: check-untyped-defs

import logging
import pickle
from dataclasses import dataclass
from functools import partial
from typing import Sequence

import jax
import jax.lax
import jax.numpy as jnp
import jax.random
import jax.tree_util
from tqdm import tqdm

from liesel.option import Option

from .chain import Chain, EpochChainManager, ListChain
from .epoch import EpochConfig, EpochManager, EpochState, EpochType
from .kernel_sequence import KernelSequence, KernelStates, TransitionInfos, TuningInfos
from .pytree import as_strong_pytree, register_dataclass_as_pytree
from .types import (
    Array,
    GeneratedQuantity,
    KeyArray,
    ModelInterface,
    ModelState,
    Position,
    PyTree,
    QuantityGenerator,
    TransitionInfo,
)

logger = logging.getLogger(__name__)


def _expand_and_stack(chunk, *rest):
    chunks = [chunk]
    chunks.extend(rest)
    expended_chunks = [jnp.expand_dims(chunk, 0) for chunk in chunks]
    return jnp.concatenate(expended_chunks, axis=0)


def stack_for_multi(chunks: list):
    """
    Combine identically structured pytrees to be used in multi-chain.

    The function adds a new dimension (axis 0) to each leaf and stacks the leafs
    along the new axis.

    This is deprecated. Please use the functions in the `pytree` module.
    """
    return jax.tree_util.tree_map(
        lambda x, *xs: _expand_and_stack(x, *xs), chunks[0], *chunks[1:]
    )


@partial(jax.jit, static_argnums=1)
def _split_keys(keys, n):
    keys = jax.lax.map(lambda key: jax.random.split(key, n), keys)
    return keys


def _initialze_prng(seed: int | KeyArray) -> KeyArray:
    if jnp.isscalar(seed):
        return jax.random.PRNGKey(seed)  # type: ignore
    elif jnp.shape(seed) == (2,):  # type: ignore
        return seed  # type: ignore
    else:
        raise ValueError("Seed has an unsupported shape")


def _add_time_dimension(multichain: bool, x: PyTree) -> PyTree:
    """
    Adds a new dimension for time to each leaf.

    The returned tree has the same structure with one additional dimension of
    size 1. The new dimension is `axis=0` if multi-chain is false and `axis=1` if
    multi-chain is true.

    If multi-chain is true, each leaf must have at least one dimension
    (representing the chain index).
    """
    axis = jax.lax.cond(multichain, None, lambda _: 1, None, lambda _: 0)
    initial_position = jax.tree_util.tree_map(
        lambda y, *_ys: jnp.expand_dims(y, axis),
        x,
    )
    return initial_position


@register_dataclass_as_pytree
@dataclass(frozen=True)
class Carry:
    kernel_states: KernelStates
    model_state: ModelState
    epoch: EpochState


@dataclass
class SamplingResult:
    """
    Contains the results of the MCMC engine.

    Easy access to the samples is provided via the methods
    `get_samples` and `get_posterior_samples`.
    """

    positions: EpochChainManager
    transition_infos: EpochChainManager
    generated_quantities: Option[EpochChainManager]
    tuning_infos: Option[Chain]
    kernel_states: Option[EpochChainManager]
    full_model_states: Option[EpochChainManager]

    def get_samples(self) -> Position:
        opt: Option[Position] = self.positions.combine_all()
        return opt.expect(f"No samples in {repr(self)}")

    def get_posterior_samples(self) -> Position:
        opt = self.positions.combine_filtered(
            lambda config: config.type == EpochType.POSTERIOR
        )
        return opt.expect(f"No posterior samples in {repr(self)}")

    def get_posterior_transition_infos(self) -> dict[str, TransitionInfo]:
        opt = self.transition_infos.combine_filtered(
            lambda config: config.type == EpochType.POSTERIOR
        )
        return opt.expect(f"No posterior transition infos in {repr(self)}")

    def get_tuning_times(self) -> Option[Array]:
        if self.tuning_infos.is_none():
            return Option.none()

        # opt_tis is not None since self.tuning_infos is not None
        opt_tis = self.tuning_infos.unwrap().get().unwrap()

        time: Array = next(iter(opt_tis.values())).time

        return Option(time)

    def pkl_save(self, path) -> None:
        """Save result as a pickled object under `path`."""
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def pkl_load(path) -> "SamplingResult":
        """Loads the pickled object from `path`."""
        with open(path, "rb") as f:
            return pickle.load(f)


class Engine:
    """MCMC engine capable of combining multiple transition kernels."""

    # __slots__ = (
    #     "_prng_keys",
    #     "_model_states",
    #     "_kernel_states",
    #     "_calculator",
    #     "_random_calculator",
    #     "_position_chain",
    #     "_transition_info_chain",
    #     "_tuning_info_chain",
    #     "_calc_result_chain",
    #     "_calc_result_random_chain",
    #     "_epoch_configs",
    #     "_position_keys",
    # )

    def __init__(
        self,
        seeds: KeyArray,
        model_states: ModelState,
        kernel_sequence: KernelSequence,
        epoch_configs: Sequence[EpochConfig],
        jitted_sample_duration: int,
        model: ModelInterface,
        position_keys: Sequence[str] | None,
        minimize_transition_infos: bool = False,
        store_kernel_states: bool = False,
        quantity_generators: Sequence[QuantityGenerator] = [],
        show_progress: bool = True,
    ):
        # fill slots that can be filled directly
        self._inital_states = model_states
        self._seeds = seeds
        self._jitted_sample_duration = jitted_sample_duration
        self._minimize_transition_infos = minimize_transition_infos
        self._store_kernel_states = store_kernel_states
        self._model_states = model_states
        self._quantity_generators = quantity_generators
        self._show_progress = show_progress

        self._kernel_sequence = kernel_sequence
        self._epoch_manager = EpochManager(epoch_configs)
        self._warmup_has_ended = False

        if not position_keys:
            position_keys = [
                key
                for ker in self._kernel_sequence._kernels  # FIXME: use of private field
                for key in ker.position_keys
            ]

        self._position_keys = position_keys
        self._model = model

        # feed in history if at least one kernel requires history for tuning
        #
        # FIXME: automatically fetch position keys
        #
        # fetch kernels' position keys and add them automatically to track them
        # in the position chain
        self._history_required_for_tuning = any(
            [ker.needs_history for ker in self._kernel_sequence._kernels]
        )  # FIXME: use of private field

        self._prng_key = seeds

        # setup storage
        self._position_chain: EpochChainManager = EpochChainManager(multichain=True)
        self._transition_info_chain: EpochChainManager = EpochChainManager(
            multichain=True
        )
        self._tuning_info_chain: ListChain = ListChain(multichain=True)
        self._kernel_state_chain: EpochChainManager = EpochChainManager(multichain=True)
        self._quantities_chain: EpochChainManager = EpochChainManager(multichain=True)

        # initialize kernel state
        keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(self._kernel_sequence.init_states)(
            keys, self._model_states
        )

        # current epoch
        self._epoch: EpochState | None = None

        # jit sample function
        self._sample_many_jitted = jax.jit(
            jax.vmap(
                self._sample_many,
                in_axes=(0, None, 0, 0),
                out_axes=(None, 0, 0, 0, 0, 0, 0),
            )
        )

    @property
    def current_epoch(self) -> EpochState:
        """
        Returns the current epoch.

        Raises a `RuntimeError` if no epoch is active.
        """
        if self._epoch is None:
            raise RuntimeError("No active epoch")

        return self._epoch

    def sample_all_epochs(self):
        """
        Runs sampling for all remaining epochs.

        Auto-tuning methods are called automatically.
        """
        while self._epoch_manager.has_more():
            self.sample_next_epoch()

    def sample_next_epoch(self):
        """Runs sampling for the next epoch assuming no epoch is active."""
        self._start_epoch()

        # special treatment for the initial values
        if self.current_epoch.config.type == EpochType.INITIAL_VALUES:
            self._handle_inital_values_epoch()
            return

        self._kernel_start_epoch()

        duration = self.current_epoch.config.duration
        epoch_type = self.current_epoch.config.type.name
        jitted = self._jitted_sample_duration

        if self._show_progress:
            logger.info(
                f"Starting epoch: {epoch_type}, {duration} transitions, "
                f"{jitted} jitted together"
            )

        self._sample_for_duration(duration=duration)
        self._end_epoch()

    def append_epoch(self, epoch: EpochConfig):
        """Appends an epoch to the epochs that should be sampled."""
        self._epoch_manager.append(epoch)

    def is_sampling_done(self) -> bool:
        """Returns true if all configured epochs have been sampled."""
        return not self._epoch_manager.has_more()

    def get_results(self) -> SamplingResult:
        """Returns the results of the sampling process."""
        if self._store_kernel_states:
            ksc = self._kernel_state_chain
        else:
            ksc = None

        if self._quantity_generators:
            gqs = self._quantities_chain
        else:
            gqs = None

        return SamplingResult(
            positions=self._position_chain,
            transition_infos=self._transition_info_chain,
            generated_quantities=Option(gqs),
            tuning_infos=Option(self._tuning_info_chain),
            kernel_states=Option(ksc),
            full_model_states=Option(None),
        )

    def _split_prng_key(self, n: int = 1) -> KeyArray:
        keys = _split_keys(self._prng_key, n + 1)
        self._prng_key = keys[:, 0, :]
        return keys[:, 1:, :]

    def _split_prng_key_one(self) -> KeyArray:
        key = self._split_prng_key(1)
        return key[:, 0, :]

    def _generate_quantity(self):
        if not self._quantity_generators:
            return None

        quants = {}

        for qg in self._quantity_generators:
            key = self._split_prng_key_one()
            gen_f = jax.vmap(qg.generate, in_axes=(0, 0, None))
            quant = gen_f(key, self._model_states, self.current_epoch)
            quants[qg.identifier] = quant

        return quants

    def _handle_inital_values_epoch(self):
        assert self.current_epoch.config.type == EpochType.INITIAL_VALUES
        self.current_epoch.advance_time(1)

        initial_position = _add_time_dimension(
            multichain=True,
            x=jax.vmap(self._model.extract_position, in_axes=(None, 0))(
                self._position_keys, self._model_states
            ),
        )

        self._position_chain.append(initial_position)

        if self._store_kernel_states:
            ks = _add_time_dimension(multichain=True, x=self._kernel_states)
            self._kernel_state_chain.append(ks)

        if self._quantity_generators:
            quants = self._generate_quantity()
            quants = _add_time_dimension(multichain=True, x=quants)
            self._quantities_chain.append(quants)

        self._epoch = None

    def _start_epoch(self):
        """Advances to the next epoch."""
        if self._epoch is not None:
            raise RuntimeError("Epoch is active and not completed")

        self._epoch = self._epoch_manager.next()

        # invoke end_warmup() for the first non-warmup epoch
        if (
            not self._warmup_has_ended
            and self.current_epoch.config.type == EpochType.POSTERIOR
        ):
            self._end_warmup()

        # advance chains to next epoch
        self._position_chain.advance_epoch(self.current_epoch.config)
        self._transition_info_chain.advance_epoch(self.current_epoch.config)
        self._kernel_state_chain.advance_epoch(self.current_epoch.config)
        self._quantities_chain.advance_epoch(self.current_epoch.config)

    def _kernel_start_epoch(self):
        """Inform kernels about new epoch."""
        keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(
            self._kernel_sequence.start_epoch, in_axes=(0, 0, 0, None)
        )(keys, self._kernel_states, self._model_states, self.current_epoch)

    def _end_warmup(self):
        """
        Ends the warmup sequence.

        Calls `end_warmup` for each kernel. From now on, only epochs of type
        posterior can follow.
        """
        keys = self._split_prng_key_one()
        tuning_infos: TuningInfos | None = self._tuning_info_chain.get().value

        end_warmup_output = jax.vmap(self._kernel_sequence.end_warmup)(
            keys, self._kernel_states, self._model_states, tuning_infos
        )
        self._kernel_states = end_warmup_output.kernel_states

        # add warnings for the user if there are any non-zero error-code
        for kid, ec in end_warmup_output.error_codes.items():
            if jnp.any(ec != 0):
                logger.warning(f"Warmup error code for {kid}: {ec}")

        logger.info("Finished warmup")

    def _end_epoch(self):
        """
        End epoch.

        Informs kernels about the end of the epoch and initializes the tuning
        if required.
        """
        # ensure that an epoch is active
        epoch = self.current_epoch

        # inform kernels about end of epoch
        end_keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(
            self._kernel_sequence.end_epoch, in_axes=(0, 0, 0, None)
        )(end_keys, self._kernel_states, self._model_states, epoch)

        self._tune_kernels(epoch)

        if self._show_progress:
            ti_option = self._transition_info_chain.get_current_chain().get()

            def count_non_zero_error_codes(tis: TransitionInfos):
                cts = {}
                for kernel_id, ti in tis.items():
                    nzero = jnp.sum(ti.error_code != 0, axis=1)
                    ntrans = ti.error_code.shape[1]  # type: ignore
                    cts[kernel_id] = (nzero, ntrans)
                return cts

            error_info: dict[str, tuple[Array, int]] = ti_option.map_or(
                {}, count_non_zero_error_codes
            )

            for kid, kcts in error_info.items():
                if jnp.any(kcts[0] != 0):
                    logger.warning(
                        f"Errors per chain for {kid}: "
                        f"{', '.join(map(str, kcts[0]))} / {kcts[1]} transitions"
                    )

            logger.info("Finished epoch")

        # no epoch is active anymore
        self._epoch = None

    def _tune_kernels(self, epoch: EpochState):
        """Trigger tuning if epoch is an adaptation phase."""
        if EpochType.is_adaptation(epoch.config.type):
            tune_keys = self._split_prng_key_one()
            if self._history_required_for_tuning:
                history = (
                    self._position_chain.get_current_chain()
                    .get()
                    .expect("The history must contain samples.")
                )
            else:
                history = None

            tune_output = jax.vmap(
                self._kernel_sequence.tune, in_axes=(0, 0, 0, None, 0)
            )(tune_keys, self._kernel_states, self._model_states, epoch, history)
            self._kernel_states = tune_output.kernel_states

            # we need to add the time dimension
            self._tuning_info_chain.append(
                _add_time_dimension(multichain=True, x=tune_output.infos)
            )

    def _sample_many(
        self,
        keys: KeyArray,
        epoch: EpochState,
        kernel_states: KernelStates,
        model_state: ModelState,
    ) -> tuple[
        EpochState,
        KernelStates,
        ModelState,
        Position,
        TransitionInfos,
        None | KernelStates,
        None | dict[str, GeneratedQuantity],
    ]:
        def scan_f(
            carry: Carry, key: KeyArray
        ) -> tuple[
            Carry,
            tuple[
                Position,
                TransitionInfos,
                None | KernelStates,
                None | dict[str, GeneratedQuantity],
            ],
        ]:
            key_trans, key_quants = jax.random.split(key)
            epoch = carry.epoch
            out = self._kernel_sequence.transition(
                key_trans, carry.kernel_states, carry.model_state, epoch
            )
            epoch.advance_time(1)
            new_carry = Carry(out.kernel_states, out.model_state, epoch)

            # extract the position specified to store in chain
            position = self._model.extract_position(
                self._position_keys, out.model_state
            )

            # minimize transition infos if requested
            tinfos = out.infos
            if self._minimize_transition_infos:
                for id in tinfos.keys():
                    tinfos[id] = tinfos[id].minimize()

            ks = None
            if self._store_kernel_states:
                ks = new_carry.kernel_states

            quants = None
            if self._quantity_generators:
                quants = {}
                keys = jax.random.split(key_quants, len(self._quantity_generators))
                for i, qg in enumerate(self._quantity_generators):
                    key = keys[i]
                    quant = qg.generate(key, out.model_state, epoch)
                    quants[qg.identifier] = quant

            return new_carry, (position, tinfos, ks, quants)

        inital_carry = Carry(kernel_states, model_state, epoch)
        carry, chain = jax.lax.scan(scan_f, inital_carry, keys)
        kernel_states = carry.kernel_states
        model_state = carry.model_state
        epoch = carry.epoch

        return (
            epoch,
            kernel_states,
            model_state,
            chain[0],
            chain[1],
            chain[2],
            chain[3],
        )

    def _sample_for_duration(self, duration: int):
        if self.current_epoch.time_left() < duration:
            raise RuntimeError("Not enough time left in epoch")

        if duration % self._jitted_sample_duration:
            raise RuntimeError(
                f"Duration {duration} is not a multiple of the "
                f"jitted sampling duration {self._jitted_sample_duration}"
            )

        # convert to non-weak device arrays to avoid recompilation
        self._epoch = as_strong_pytree(self._epoch)
        self._kernel_states = as_strong_pytree(self._kernel_states)
        self._model_states = as_strong_pytree(self._model_states)

        it = range(duration // self._jitted_sample_duration)

        if self._show_progress:
            it = tqdm(it, ncols=80, disable=None, unit="chunk")

        for dur_i in it:
            # FIXME: split for entire duration instead of each loop iteration
            keys = self._split_prng_key(self._jitted_sample_duration)
            (
                new_epoch,
                new_ks,
                new_ms,
                position_chain,
                infos,
                ksc,
                quants,
            ) = self._sample_many_jitted(
                keys, self.current_epoch, self._kernel_states, self._model_states
            )

            self._epoch = new_epoch
            self._kernel_states = new_ks
            self._model_states = new_ms
            self._position_chain.append(position_chain)
            self._transition_info_chain.append(infos)
            if self._store_kernel_states:
                self._kernel_state_chain.append(ksc)
            if self._quantity_generators:
                self._quantities_chain.append(quants)



"""
# The Liesel-Goose interface
"""

from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, Callable, Iterable

from .types import ModelState, Position

if TYPE_CHECKING:
    from .model import Model


def get_position(position_keys: Iterable[str], model_state: ModelState) -> Position:
    """
    Extracts a position from a model state.

    A position is a dictionary of node names and values.
    """

    return Position({key: model_state[key].value for key in position_keys})


def update_state(
    position: Position, model_state: ModelState, model: Model
) -> ModelState:
    """Updates and returns a model state given a position."""

    model.state = model_state

    for node in model.sorted_nodes:
        node.outdated = False

    for name, value in position.items():
        model.nodes[name].set_value(value, update=False)

    model.update()
    return model.state


def get_log_prob(model_state: ModelState) -> float:
    """Computes and returns the log-probability from a model state."""

    log_probs = [node_state.log_prob for node_state in model_state.values()]
    return reduce(lambda x, y: x + y, log_probs)


def make_update_state_fn(
    model: Model, jaxify=True
) -> Callable[[Position, ModelState], ModelState]:
    """Returns a pure and jittable `update_state` function for the provided model."""

    model = model.empty_copy()

    if jaxify:
        model.jaxify()

    def fn(position: Position, model_state: ModelState) -> ModelState:
        return update_state(position, model_state, model)

    return fn


def make_log_prob_fn(
    model: Model, jaxify=True
) -> Callable[[Position, ModelState], float]:
    """Returns a pure and jittable `log_prob` function for the provided model."""

    model = model.empty_copy()

    if jaxify:
        model.jaxify()

    def fn(position: Position, model_state: ModelState) -> float:
        return get_log_prob(update_state(position, model_state, model))

    return fn


class GooseModel:
    """A `liesel.goose.ModelInterface` for a `liesel.liesel.Model`."""

    def __init__(self, model: Model):
        self._update_state_fn = make_update_state_fn(model)

    def extract_position(
        self, position_keys: Iterable[str], model_state: ModelState
    ) -> Position:
        return get_position(position_keys, model_state)

    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        return self._update_state_fn(position, model_state)

    def log_prob(self, model_state: ModelState) -> float:
        return get_log_prob(model_state)



"""
# Metropolis-Hastings

This module uses the error codes 90-99.
"""

import jax
import jax.numpy as jnp

from .kernel import DefaultTransitionInfo
from .types import KeyArray, ModelInterface, ModelState, Position

mh_error_book = {0: "no errors", 90: "nan acceptance prob"}
"""The error book of the `mh_step` function."""


def mh_step(
    prng_key: KeyArray,
    model: ModelInterface,
    proposal: Position,
    model_state: ModelState,
    correction: float = 0.0,
) -> tuple[DefaultTransitionInfo, ModelState]:
    """
    Decides if an MCMC proposal is accepted in a Metropolis-Hastings step.

    ## Parameters

    - `correction`: The Metropolis-Hastings correction in the case of an asymmetric
      proposal distribution.

    ## Returns

    A tuple of a `liesel.goose.types.TransitionInfo` and a
    `liesel.goose.types.ModelState` (= a pytree).
    """

    current_log_prob = model.log_prob(model_state)
    proposed_model_state = model.update_state(proposal, model_state)
    proposed_log_prob = model.log_prob(proposed_model_state)

    log_acc_prob = proposed_log_prob - current_log_prob + correction

    log_acc_prob, error_code = jax.lax.cond(
        jnp.isnan(log_acc_prob),
        lambda: (-jnp.inf, 90),
        lambda: (log_acc_prob, 0),
    )

    acceptance_prob = jnp.clip(jnp.exp(log_acc_prob), a_max=1.0)
    do_accept = jax.random.uniform(prng_key) <= acceptance_prob

    info = DefaultTransitionInfo(error_code, acceptance_prob, do_accept)

    model_state = jax.lax.cond(
        do_accept,
        lambda: proposed_model_state,
        lambda: model_state,
    )

    return info, model_state



# GENERATED FILE, DO NOT EDIT!

"""
# An algebraic sigmoid bijector
"""

import jax.numpy as np
import tensorflow_probability.substrates.jax.bijectors as tfb

__docformat__ = "google"


class AlgebraicSigmoid(tfb.Bijector):
    """
    The algebraic sigmoid bijector `f(x) = x / sqrt(1 + x^2)`.
    """

    def __init__(self, validate_args=False, name="algebraic_sigmoid"):
        parameters = dict(locals())

        super().__init__(
            forward_min_event_ndims=0,
            validate_args=validate_args,
            parameters=parameters,
            name=name,
        )

    def _forward(self, x):
        return x / np.sqrt(1.0 + x**2)

    def _inverse(self, y):
        return y / np.sqrt(1.0 - y**2)

    def _inverse_log_det_jacobian(self, y):
        return -1.5 * np.log(1.0 - y**2)

    def _forward_log_det_jacobian(self, x):
        return -1.5 * np.log(1.0 + x**2)

    @classmethod
    def _is_increasing(cls):
        return True



"""
# Liesel modeling framework
"""

from .distreg import CopRegBuilder, DistRegBuilder, dist_reg_mcmc, tau2_gibbs_kernel
from .goose import GooseModel
from .model import Model, ModelBuilder, load_model, save_model
from .nodes import (
    PIT,
    Addition,
    Bijector,
    ColumnStack,
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Node,
    NodeCalculator,
    NodeDistribution,
    NodeGroup,
    NodeState,
    Parameter,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
    transform_parameter,
)
from .types import ModelState, Position
from .viz import plot_model

__all__ = [
    "Addition",
    "Bijector",
    "ColumnStack",
    "CopRegBuilder",
    "DesignMatrix",
    "DistRegBuilder",
    "GooseModel",
    "Hyperparameter",
    "InverseLink",
    "Model",
    "ModelBuilder",
    "ModelState",
    "Node",
    "NodeCalculator",
    "NodeDistribution",
    "NodeGroup",
    "NodeState",
    "PIT",
    "Parameter",
    "Position",
    "Predictor",
    "RegressionCoef",
    "Response",
    "Smooth",
    "SmoothingParam",
    "dist_reg_mcmc",
    "load_model",
    "plot_model",
    "save_model",
    "tau2_gibbs_kernel",
    "transform_parameter",
]

ModelState = ModelState
"""The state of a model as a dictionary of node names and states."""

Position = Position
"""The position of a model as a dictionary of node names and values."""



"""
# Goose MCMC framework

This module defines the public API of Goose. Expect breakages if using anything
beyond this module (and maybe also if using this module, as this is early-stage
research software).
"""

from .builder import EngineBuilder as EngineBuilder
from .engine import Engine as Engine
from .epoch import EpochConfig, EpochType
from .gibbs import GibbsKernel
from .hmc import HMCKernel
from .iwls import IWLSKernel
from .models import DictModel
from .nuts import NUTSKernel
from .rw import RWKernel
from .summary_m import Summary, summary
from .summary_viz import plot_cor, plot_density, plot_param, plot_trace
from .types import ModelInterface, Position
from .warmup import stan_epochs

__all__ = [
    "DictModel",
    "Engine",
    "EngineBuilder",
    "EpochConfig",
    "EpochType",
    "GibbsKernel",
    "HMCKernel",
    "IWLSKernel",
    "ModelInterface",
    "NUTSKernel",
    "Position",
    "RWKernel",
    "Summary",
    "plot_cor",
    "plot_density",
    "plot_param",
    "plot_trace",
    "stan_epochs",
    "summary",
]



"""
# Distributional regression
"""

from __future__ import annotations

import jax.numpy as jnp
import jax.random
import numpy as np

from liesel.goose import EngineBuilder, GibbsKernel, IWLSKernel
from liesel.option import Option

from .goose import GooseModel
from .model import Model, ModelBuilder
from .nodes import (
    PIT,
    ColumnStack,
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Node,
    NodeDistribution,
    NodeGroup,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
)
from .types import Array, TFPBijectorClass, TFPDistributionClass

matrix_rank = np.linalg.matrix_rank


class SmoothGroup(NodeGroup):
    """A node group representing a smooth."""


class PSmoothGroup(SmoothGroup):
    """A node group representing a parametric smooth."""


class NPSmoothGroup(SmoothGroup):
    """A node group representing a non-parametric smooth."""


class DistRegBuilder(ModelBuilder):
    """A model builder for distributional regression models."""

    _howto = "Smooths need to be added first, then the predictors, then the response"

    def __init__(self) -> None:
        super().__init__()

        self._smooths: dict[str, list[Node]] = {}
        self._distributional_parameters: dict[str, Node] = {}
        self._response: Option[Node] = Option(None)

    @property
    def response(self) -> Node:
        return self._response.expect(f"No response in {repr(self)}")

    def _smooth_name(self, name: str | None, predictor: str, prefix: str) -> str:
        """Generates a name for a smooth if the `name` argument is `None`."""

        other_smooths = self._smooths[predictor]
        other_names = [node.name for node in other_smooths if node.has_name]
        prefix = predictor + "_" + prefix
        counter = 0

        while prefix + str(counter) in other_names:
            counter += 1

        if not name:
            name = prefix + str(counter)

        if name in other_names:
            raise RuntimeError(
                f"Smooth {repr(name)} already exists in {repr(self)} "
                f"for predictor {repr(predictor)}"
            )

        return name

    def add_p_smooth(
        self,
        X: Array,
        m: float,
        s: float,
        predictor: str,
        name: str | None = None,
    ) -> PSmoothGroup:
        """Adds a parametric smooth to the model builder.

        ## Parameters

        - `X`: The design matrix.
        - `m`: The mean of the Gaussian prior.
        - `s`: The standard deviation of the Gaussian prior.
        - `predictor`: The name of the predictor to add the smooth to.
        - `name`: The name of the smooth.
        """

        try:
            self._smooths[predictor]
        except KeyError:
            self._smooths[predictor] = []

        name = self._smooth_name(name, predictor, "p")

        X_node = DesignMatrix(X, name=name + "_X")
        m_node = Hyperparameter(m, name=name + "_m")
        s_node = Hyperparameter(s, name=name + "_s")

        beta = np.zeros(np.shape(X)[-1], np.float32)
        beta_distribution = NodeDistribution("Normal", loc=m_node, scale=s_node)
        beta_node = RegressionCoef(beta, beta_distribution, name + "_beta")

        smooth_node = Smooth(X_node, beta_node, name=name)
        self._smooths[predictor].append(smooth_node)

        group = PSmoothGroup(
            smooth=smooth_node,
            beta=beta_node,
            X=X_node,
            m=m_node,
            s=s_node,
        )

        group.name = name
        self.add_groups(group)
        return group

    def add_np_smooth(
        self,
        X: Array,
        K: Array,
        a: float,
        b: float,
        predictor: str,
        name: str | None = None,
    ) -> NPSmoothGroup:
        """
        Adds a non-parametric smooth to the model builder.

        ## Parameters

        - `X`: The design matrix.
        - `K`: The penalty matrix.
        - `a`: The a, α or concentration parameter of the inverse gamma prior.
        - `b`: The b, β or scale parameter of the inverse gamma prior.
        - `predictor`: The name of the predictor to add the smooth to.
        - `name`: The name of the smooth.
        """

        try:
            self._smooths[predictor]
        except KeyError:
            self._smooths[predictor] = []

        name = self._smooth_name(name, predictor, "np")

        X_node = DesignMatrix(X, name=name + "_X")
        K_node = Hyperparameter(K, name=name + "_K")
        a_node = Hyperparameter(a, name=name + "_a")
        b_node = Hyperparameter(b, name=name + "_b")

        rank_node = Hyperparameter(matrix_rank(K), name=name + "_rank")

        tau2_parameters = {"concentration": a_node, "scale": b_node}
        tau2_distribution = NodeDistribution("InverseGamma", **tau2_parameters)
        tau2_node = SmoothingParam(10000.0, tau2_distribution, name + "_tau2")

        beta = np.zeros(np.shape(X)[-1], np.float32)
        beta_parameters = {"tau2": tau2_node, "K": K_node, "rank": rank_node}
        beta_distribution = NodeDistribution("SmoothPrior", **beta_parameters)
        beta_node = RegressionCoef(beta, beta_distribution, name + "_beta")

        smooth_node = Smooth(X_node, beta_node, name=name)
        self._smooths[predictor].append(smooth_node)

        group = NPSmoothGroup(
            smooth=smooth_node,
            beta=beta_node,
            tau2=tau2_node,
            rank=rank_node,
            X=X_node,
            K=K_node,
            a=a_node,
            b=b_node,
        )

        group.name = name
        self.add_groups(group)
        return group

    def add_predictor(
        self, name: str, inverse_link: str | TFPBijectorClass
    ) -> DistRegBuilder:
        """
        Adds a predictor to the model builder.

        ## Parameters

        - `name`:
          The name of the parameter of the response distribution.

          Must match the name of the parameter of the TFP distribution.

        - `inverse_link`:
          The inverse link mapping the regression predictor to the parameter
          of the response distribution.

          Either a string identifying a TFP bijector, or alternatively,
          a TFP-compatible bijector class. If a class is provided instead of a string,
          the user needs to make sure it uses the right NumPy implementation.
        """

        if name not in self._smooths or not self._smooths[name]:
            msg = f"No smooths in {repr(self)} for predictor {repr(name)}. "
            raise RuntimeError(msg + self._howto)

        smooth_nodes = self._smooths[name]
        predictor_node = Predictor(name=name + "_pdt", *smooth_nodes)
        parameter_node = InverseLink(inverse_link, predictor_node, name=name)
        self._distributional_parameters[name] = parameter_node
        self.add_nodes(predictor_node, parameter_node)

        return self

    def add_response(
        self, response: Array, distribution: str | TFPDistributionClass
    ) -> DistRegBuilder:
        """
        Adds the response to the model builder.

        ## Parameters

        - `response`:
          The response vector or matrix.

        - `distribution`:
          The conditional distribution of the response variable.

          Either a string identifying a TFP distribution, or alternatively,
          a TFP-compatible distribution class. If a class is provided instead of
          a string, the user needs to make sure it uses the right NumPy implementation.
        """

        if not self._distributional_parameters:
            raise RuntimeError(f"No predictors in {repr(self)}. {self._howto}")

        response_distribution = NodeDistribution(
            distribution, **self._distributional_parameters
        )

        response_node = Response(response, response_distribution, "response")
        self._response = Option(response_node)
        self.add_nodes(response_node)

        return self


class CopRegBuilder(DistRegBuilder):
    """
    A model builder for copula regression models.

    Remember to add a predictor for the dependence parameter before adding the copula
    and building the model.

    ## Parameters

    - `model0`: The first marginal distributional regression model **builder**.
    - `model1`: The second marginal distributional regression model **builder**.
    - `copula`: The copula of the response variables.

      Either a string identifying a TFP distribution, or alternatively,
      a TFP-compatible distribution class. If a class is provided instead of a string,
      the user needs to make sure it uses the right NumPy implementation.
    """

    def __init__(self, model0: DistRegBuilder, model1: DistRegBuilder) -> None:
        super().__init__()

        arg0_is_mb = isinstance(model0, DistRegBuilder)
        arg1_is_mb = isinstance(model1, DistRegBuilder)

        if not arg0_is_mb or not arg1_is_mb:
            raise RuntimeError(f"Arguments of {repr(self)} must be model builders")

        self.groups = model0.groups + model1.groups
        self.nodes = model0.nodes + model1.nodes

        self._model0 = self._update_names(model0, "m0_")
        self._model1 = self._update_names(model1, "m1_")

    @staticmethod
    def _update_names(model: DistRegBuilder, prefix: str) -> DistRegBuilder:
        for node in model.all_nodes():
            node.name = prefix + node.name

        for group in model.groups:
            group.name = prefix + group.name

        return model

    def add_copula(self, copula: str | TFPDistributionClass) -> CopRegBuilder:
        pit0_node = PIT(self._model0.response, name="m0_pit")
        pit1_node = PIT(self._model1.response, name="m1_pit")

        copula_node = ColumnStack(
            pit0_node,
            pit1_node,
            distribution=NodeDistribution(copula, **self._distributional_parameters),
            name="copula",
        )

        self._response = Option(copula_node)
        self.add_nodes(pit0_node, pit1_node, copula_node)
        return self


def tau2_gibbs_kernel(group: NPSmoothGroup) -> GibbsKernel:
    """Builds a Gibbs kernel for a smoothing parameter with an inverse gamma prior."""

    position_key = group["tau2"].name

    def transition(prng_key, model_state):
        a_prior = model_state[group["a"].name].value
        rank = model_state[group["rank"].name].value

        a_gibbs = jnp.squeeze(a_prior + 0.5 * rank)

        b_prior = model_state[group["b"].name].value
        beta = model_state[group["beta"].name].value
        K = model_state[group["K"].name].value

        b_gibbs = jnp.squeeze(b_prior + 0.5 * (beta @ K @ beta))

        draw = b_gibbs / jax.random.gamma(prng_key, a_gibbs)

        return {position_key: draw}

    return GibbsKernel([position_key], transition)


def dist_reg_mcmc(model: Model, seed: int, num_chains: int) -> EngineBuilder:
    """
    Configures an `EngineBuilder` with a Metropolis-in-Gibbs MCMC algorithm with an
    `IWLSKernel` for the regression coefficients and a `GibbsKernel` for the smoothing
    parameters for a distributional regression model.

    ## Parameters

    - `model`: A model built with a `DistRegBuilder`.
    - `seed`: The PRNG seed for the engine builder.
    - `num_chains`: The number of chains to be sampled.
    """

    builder = EngineBuilder(seed, num_chains)

    builder.set_model(GooseModel(model))
    builder.set_initial_values(model.state)

    for group in model.groups.values():
        if not isinstance(group, SmoothGroup):
            next

        if isinstance(group, NPSmoothGroup):
            tau2_kernel = tau2_gibbs_kernel(group)
            builder.add_kernel(tau2_kernel)

        position_key = group["beta"].name
        beta_kernel = IWLSKernel([position_key])
        builder.add_kernel(beta_kernel)

    return builder



"""
# Pytree utilities
"""

import dataclasses
from typing import TypeVar

import jax
import jax.numpy as jnp
import jax.tree_util

T = TypeVar("T")


def register_dataclass_as_pytree(cls):
    """Decorator for registering dataclasses as pytrees."""

    if not dataclasses.is_dataclass(cls):
        raise TypeError(f"{cls} must be a dataclass")

    def flatten(cls_instance):
        # don't use dataclasses.asdict() here, because it converts nested dataclasses
        # to dicts recursively

        return jax.tree_util.tree_flatten(cls_instance.__dict__)

    def unflatten(aux_data, children):
        d = jax.tree_util.tree_unflatten(aux_data, children)
        rv = cls.__new__(cls)
        rv.__dict__.update(d)
        return rv

    jax.tree_util.register_pytree_node(cls, flatten, unflatten)

    return cls


def slice_leaves(pytree, idx):
    """
    Performs the same slice operation on every leaf.

    `idx` can be constructed with `jnp.s_` or `np.s_`, for example:

    ```python
    jnp.s_[0]
    jnp.s_[0:3, :, 2]
    ```
    """
    return jax.tree_util.tree_map(lambda x: x[idx], pytree)


def stack_leaves(pytrees, axis=0):
    """
    Stacks all leaves in the list of pytrees along the given axis.

    The stack operation creates a new axis.
    """
    return jax.tree_util.tree_map(
        lambda *xs: jnp.stack(xs, axis=axis),
        *pytrees,
    )


def concatenate_leaves(pytrees, axis=0):
    """
    Concatenates all leaves in the list of pytrees along the given axis.

    The concatenate operation does not create a new axis.
    """
    return jax.tree_util.tree_map(lambda *xs: jnp.concatenate(xs, axis=axis), *pytrees)


def split_leaves(pytree, indices_or_sections, axis=0):
    """
    Splits all leaves in a pytree into multiple sub-arrays.

    The function applies `jnp.split` to all leaves.
    """
    return jax.tree_util.tree_map(
        lambda x: jnp.split(x, indices_or_sections, axis), pytree
    )


def squeeze_leaves(pytree, axis=0):
    """
    Squeezes all leaves in a pytree.
    """
    return jax.tree_util.tree_map(lambda x: jnp.squeeze(x, axis), pytree)


def split_and_transpose(pytree, axis=0):
    """
    Splits the leaves in a pytree along one axis and transposes the tree such
    that it's a list of pytrees. It assumes that all leaves have the same
    dimensionality along the chosen axis.
    """
    dim = jax.tree_util.tree_leaves(pytree)[0].shape[axis]
    spytree = split_leaves(pytree, dim, axis=axis)
    td_inner = jax.tree_util.tree_structure([0 for _ in range(dim)])
    td_outer = jax.tree_util.tree_structure(pytree)
    return jax.tree_util.tree_transpose(td_outer, td_inner, spytree)


def as_strong_pytree(pytree: T) -> T:
    """
    Converts every leaf in a pytree to a non-weak `DeviceArray`.

    See <https://jax.readthedocs.io/en/latest/type_promotion.html>.
    """
    return jax.tree_util.tree_map(
        lambda x: jnp.asarray(x, dtype=jnp.asarray(x).dtype), pytree
    )



# GENERATED FILE, DO NOT EDIT!

"""
# A normal smooth prior for semi-parametric regression
"""

import jax.numpy as np
import tensorflow_probability.substrates.jax.distributions as tfd
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.substrates.jax import tf2jax as tf

__docformat__ = "google"


class SmoothPrior(tfd.Distribution):
    """
    A (potentially rank-deficient) normal smooth prior for semi-parametric regression.

    See Ludwig Fahrmeir et al., [Regression (2013), Section 8.1.3](
    https://link.springer.com/book/10.1007/978-3-642-34333-9) for details.
    """

    def __init__(
        self,
        tau2,
        K,
        rank=None,
        validate_args=False,
        allow_nan_stats=True,
        name="SmoothPrior",
    ):
        """
        Constructs the distribution.

        Args:
            tau2: The smoothing parameter.
            K: The (potentially rank-deficient) penalty matrix.
            rank: The rank of the penalty matrix. Computed from `K` if not provided.
        """

        parameters = dict(locals())

        if rank is None:
            event_shape = np.shape(K)[-1]
            signature = f"({event_shape},{event_shape})->()"
            rank_fn = np.vectorize(np.linalg.matrix_rank, signature=signature)
            rank = rank_fn(K)

        self._rank = rank
        self._tau2 = tau2
        self._K = K

        super().__init__(
            dtype=K.dtype,
            reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
            parameters=parameters,
            name=name,
        )

    def _log_prob(self, x):
        x = np.expand_dims(x, axis=-2)
        x_T = np.swapaxes(x, -2, -1)

        p1 = -0.5 * self._rank * np.log(self._tau2)
        p2 = -0.5 / self._tau2 * np.squeeze(x @ self._K @ x_T, axis=(-2, -1))
        return p1 + p2

    def _event_shape(self):
        return tf.TensorShape((np.shape(self._K)[-1],))

    def _event_shape_tensor(self):
        return np.array((np.shape(self._K)[-1],), dtype=np.int32)

    def _batch_shape(self):
        return tf.TensorShape(np.shape(self._tau2))

    def _batch_shape_tensor(self):
        return np.array(np.shape(self._tau2), dtype=np.int32)



"""
# Type aliases, type variables and protocols
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, NewType, Protocol, Sequence, TypeVar

import jax.numpy

if TYPE_CHECKING:
    from .epoch import EpochState
    from .kernel import (
        DefaultTransitionInfo,
        TransitionOutcome,
        TuningOutcome,
        WarmupOutcome,
    )


# simple type aliases

PyTree = Any
Array = Any

ModelState = PyTree
Position = NewType("Position", dict[str, Array])
KernelState = PyTree
KeyArray = jax.random.KeyArray


class TuningInfo(Protocol):
    error_code: int
    time: int


class TransitionInfo(Protocol):
    error_code: int
    """An error code defined in the error book of the kernel.
    0 if no errors occurred during the transition."""

    acceptance_prob: float
    """The acceptance probability of a proposal of a Metropolis-Hastings kernel
    or the average acceptance probability across a trajectory of a NUTS-type kernel.
    99.0 if not used by the kernel."""

    position_moved: int
    """0 if the position did not move during the transition, 1 if it *did* move,
    and 99 if unknown."""

    @abstractmethod
    def minimize(self) -> DefaultTransitionInfo:
        raise NotImplementedError


TKernelState = TypeVar("TKernelState", bound=KernelState)
TTransitionInfo = TypeVar("TTransitionInfo", bound=TransitionInfo)
TTuningInfo = TypeVar("TTuningInfo", bound=TuningInfo)


class ModelInterface(Protocol):
    """
    The model interface defines a standardized way for Goose to communicate
    with a statistical model, that is, to update the model state and to compute
    the log-probability.
    """

    @abstractmethod
    def extract_position(
        self, position_keys: Sequence[str], model_state: ModelState
    ) -> Position:
        """
        Extracts the position from the model state given a sequence of position keys.

        The returned position must have the same order as the supplied key sequence.
        """

        raise NotImplementedError

    @abstractmethod
    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        """Updates the model state with the values in the position."""

        raise NotImplementedError

    @abstractmethod
    def log_prob(self, model_state: ModelState) -> float:
        """Computes the unnormalized log-probability given the model state."""

        raise NotImplementedError


class Kernel(Protocol[TKernelState, TTransitionInfo, TTuningInfo]):
    """Protocol for a transition kernel."""

    error_book: ClassVar[dict[int, str]]
    """
    Maps error codes to error messages.

    We use -1 if an error code is required but the kernel is skipped.
    """

    needs_history: ClassVar[bool] = False
    """Is set to true if the kernel expects the history for tuning."""

    position_keys: tuple[str, ...]
    """Keys for which the kernel handles the transition."""

    identifier: str = ""
    """
    An identifier for the kernel object that is set by the `EngineBuilder`
    if it is an empty string.
    """

    @abstractmethod
    def set_model(self, model: ModelInterface):
        ...

    @abstractmethod
    def has_model(self) -> bool:
        ...

    @abstractmethod
    def init_state(self, prng_key: KeyArray, model_state: ModelState) -> KernelState:
        """Creates the initial kernel state."""

        raise NotImplementedError

    @abstractmethod
    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """Handles one transition. Must be jittable."""

        raise NotImplementedError

    @abstractmethod
    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        Called after each warmup epoch.

        `history` may be `None` if class variable `needs_history` is `False`.

        Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TKernelState:
        """
        Called at the beginning of an epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TKernelState:
        """
        Called at the end of an epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        tuning_history: TTuningInfo | None,
    ) -> WarmupOutcome[TKernelState]:
        """
        Asks the kernel to inspect the warmup history and react to it.

        This method is executed once the first non-warmup epoch is encountered
        and before `start_epoch` is called.

        `tuning_history` is `None` if no tuning has happened prior to the first
        non-warmup epoch. Otherwise `tuning_history` has the same structure as returned
        from `tune()` but each leaf has an additional dimension. The first dimension
        refers to the n-the tuning.

        Must be jittable.

        The signature is likely to change.
        """

        raise NotImplementedError


class GeneratedQuantity(Protocol):
    error_code: int


TGeneratedQuantity = TypeVar(
    "TGeneratedQuantity", bound=GeneratedQuantity, covariant=True
)


class QuantityGenerator(Protocol[TGeneratedQuantity]):
    error_book: ClassVar[dict[int, str]]
    identifier: str

    @abstractmethod
    def set_model(self, model: ModelInterface):
        raise NotImplementedError

    @abstractmethod
    def has_model(self) -> bool:
        raise NotImplementedError

    def generate(
        self,
        prng_key: KeyArray,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TGeneratedQuantity:
        """Generates a new quantity based on the model and PRNG state."""



from .__version__ import __version__, __version_info__  # isort: skip  # noqa: F401

from . import goose, liesel, tfp
from .logging import setup_logger

# Because logger setup takes place after importing the submodules, it only affects
# log messages emitted at runtime.
setup_logger()

__all__ = ["goose", "liesel", "tfp"]



"""
# MCMC engine builder

The purpose of the engine builder is to provide a simple API to gradually assemble
the components needed by the MCMC engine. The builder is responsible of returning an
engine in a well-defined state. Furthermore, the builder can return different engine
implementations.
"""

import math
from collections.abc import Iterable
from typing import cast

import jax
import jax.numpy as jnp

from liesel.option import Option

from .engine import Engine
from .epoch import EpochConfig, EpochManager
from .kernel_sequence import KernelSequence
from .pytree import stack_leaves
from .types import Kernel, KeyArray, ModelInterface, ModelState, QuantityGenerator
from .warmup import stan_epochs


def _find_duplicate(xs: list[str]) -> Option[str]:
    """Checks if a list of strings contains any duplicates."""
    set_ = set()

    for x in xs:
        if x in set_:
            return Option(x)
        else:
            set_.add(x)

    return Option(None)


class EngineBuilder:
    """
    The `EngineBuilder` is used to construct an MCMC Engine.

    Currently, the `EngineBuilder` builds an object of the class `Engine`.

    By default, every position key associated with an MCMC kernel is tracked.
    This behavior can be adjusted with the fields `positions_included`
    and `positions_excluded`.

    ## Parameters

    - `seed`: Used to initialize the PRNG for the building process. The PRNG state
      is also used to initialize the PRNG state of the engine. A user-specific seed
      to initialize the engine can be set with the method `set_engine_seed`.
    - `num_chains`: The number of chains to be used.
    """

    def __init__(self, seed: int, num_chains: int):
        keys = jax.random.split(jax.random.PRNGKey(seed))
        self._prng_key: KeyArray = keys[0]
        self._engine_key: KeyArray = keys[1]
        self._num_chains: int = num_chains
        self._kernels: list[Kernel] = []
        self._quantity_generators: list[QuantityGenerator] = []
        self._model_state: Option[ModelState] = Option(None)

        self._model: Option[ModelInterface] = Option(None)

        # public fields, only simple states
        self.store_kernel_states: bool = False
        self.minimize_transition_infos: bool = False
        self.show_progress: bool = True

        self.positions_included: list[str] = []
        """List of additional position keys that should be tracked."""

        self.positions_excluded: list[str] = []
        """
        List of position keys that should not be tracked.

        Excluded keys override additional keys.
        """

    def set_engine_seed(self, seed: int | KeyArray):
        """Sets a seed used to initialize the MCMC engine."""
        if jnp.isscalar(seed):
            seed_int: int = cast(int, seed)
            self._engine_key = jax.random.PRNGKey(seed_int)
        else:
            seed_keyarray = cast(KeyArray, seed)
            self._engine_key = seed_keyarray

    @property
    def engine_seed(self) -> KeyArray:
        return self._engine_key

    def add_kernel(self, kernel: Kernel):
        """Adds a `Kernel`."""
        self._kernels.append(kernel)

    @property
    def kernels(self) -> tuple[Kernel, ...]:
        return tuple(self._kernels)

    def add_quantity_generator(self, generator: QuantityGenerator):
        """Adds a `QuantityGenerator`."""
        self._quantity_generators.append(generator)

    @property
    def quantity_generators(self) -> tuple[QuantityGenerator, ...]:
        return tuple(self._quantity_generators)

    def set_initial_values(self, model_state: ModelState, multiple_chains=False):
        """
        Sets the initial model state.

        If `multiple_chains` is true the `model_state` will be used as is;
        otherwise `model_state` will be used as the initial values for each chain.
        Note that if `multiple_chains` is true, the first axis of each leaf of
        `model_state` refers to the chain.
        """
        if not multiple_chains:
            model_states = stack_leaves(model_state for _ in range(self._num_chains))

        self._model_state = Option(model_states)

    @property
    def model_state(self) -> Option[ModelState]:
        return self._model_state

    def set_epochs(self, epochs: Iterable[EpochConfig]):
        """Sets epochs."""
        self._epochs = EpochManager(epochs)

    def set_duration(
        self, warmup_duration: int, posterior_duration: int, term_duration: int = 50
    ):
        """
        Sets epochs using the `stan_epochs` function.

        Note that `term_duration` needs to be long enough that tuning algorithms
        like dual averaging can converge.
        """
        epochs = stan_epochs(
            warmup_duration, posterior_duration, term_duration=term_duration
        )
        self._epochs = EpochManager(epochs)

    @property
    def epochs(self) -> tuple[EpochConfig, ...]:
        return tuple(self._epochs._configs)

    def set_model(self, model: ModelInterface):
        """Sets the model interface for all kernels and quantity generators."""
        self._model = Option(model)

    def build(self) -> Engine:
        """Builds the MCMC engine with the provided setup."""
        # build list of position keys
        pos_keys: list[str] = []
        for ker in self._kernels:
            pos_keys.extend(ker.position_keys)
        dupl = _find_duplicate(pos_keys)
        if dupl.is_some():
            raise RuntimeError(
                f"The position key {dupl.unwrap()} is claimed by multiple kernels"
            )

        pos_keys.extend(self.positions_included)
        pos_keys = [key for key in pos_keys if key not in self.positions_excluded]

        # find good jittable number
        epochs = self._epochs._configs  # FIXME: use of private field
        durations = [e.duration for e in epochs[1:]]
        jit_duration = math.gcd(*durations)

        # seeds
        seeds = self._engine_key
        if seeds.shape == (2,):
            # no multi-chain key
            seeds = jax.random.split(seeds, self._num_chains)
        if seeds.shape != (self._num_chains, 2):
            raise RuntimeError(
                f"MCMC seed has the wrong dimensions {seeds}. "
                f"Expected is {(self._num_chains, 2)}"
            )

        # check for duplicated identifiers in self._quantity_generators
        idents = []
        for qg in self._quantity_generators:
            idents.append(qg.identifier)

        dupl = _find_duplicate(idents)
        if dupl.is_some():
            raise RuntimeError(
                f"The identifier {dupl.unwrap()} is used by multiple "
                "quantity generators"
            )

        # set model interface for all kernels and quantity generators
        model = self._model.expect("Model interface must be set")
        for ker in self.kernels:
            if not ker.has_model():
                ker.set_model(model)
        for qg in self.quantity_generators:
            if not qg.has_model():
                qg.set_model(model)

        # assign identifiers to kernels
        for idx, ker in enumerate(self.kernels):
            if not ker.identifier:
                ker.identifier = f"kernel_{idx:02d}"

        return Engine(
            seeds=seeds,
            model_states=self._model_state.expect("Model state must be set"),
            kernel_sequence=KernelSequence(self.kernels),
            epoch_configs=epochs,
            jitted_sample_duration=jit_duration,
            model=model,
            position_keys=pos_keys,
            minimize_transition_infos=self.minimize_transition_infos,
            store_kernel_states=self.store_kernel_states,
            quantity_generators=self.quantity_generators,
            show_progress=self.show_progress,
        )



import logging
from pathlib import Path


def setup_logger() -> None:
    """
    Sets up a basic `StreamHandler`, which prints log messages to the terminal.
    The default log level of the `StreamHandler` is set to "info".
    """

    logger = logging.getLogger("root")
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)

    formatter = logging.Formatter("%(levelname)s - %(message)s")

    ch.setFormatter(formatter)
    logger.addHandler(ch)


def add_file_handler(
    path: str | Path,
    level: str,
    logger: str = "liesel",
    fmt: str = "%(asctime)s - %(levelname)s - %(name)s - %(message)s",
) -> None:
    """
    Adds a file handler for logging output.

    ## Arguments

    - `path`: Absolute path to log file. If it does not exist, it will be created.
      If any parent directory does not exist, it will be created as well.
    - `level`: The level of messages to log to the specified file. Can be "debug",
      "info", "warning", "error" or "critical". The logger will catch all messages
      from the specified level upwards.
    - `logger`: The name of the logger to configure the file handler for. Can be,
      for instance, the full name of a Liesel module. For example, to configure a
      logger for `liesel.goose`, the logger should be specified as "liesel.goose".
    - `fmt`: Formatting string, see the documentation of the standard library's
      `logging.Formatter` class.

    ## Examples

    A basic file handler, catching all log messages in the Liesel framework:

    ```python
    import logging
    import liesel as lsl

    lsl.logging.add_file_handler(path="path/to/logfile.log", level="debug")

    logger = logging.getLogger("liesel")
    logger.warning("My warning message")
    ```

    A file handler that catches only log messages from the `liesel.goose` module of
    level "warning" or higher:

    ```python
    import logging
    import liesel as lsl

    lsl.logging.add_file_handler(
        path="path/to/goose_warnings.log",
        level="warning",
        logger="liesel.goose"
    )

    logger = logging.getLogger("liesel")
    logger.warning("My warning message")
    ```
    """

    path = Path(path)

    if not path.is_absolute():
        raise ValueError("Provided path for logging file handler must be absolute")

    path.parent.mkdir(parents=True, exist_ok=True)

    log = logging.getLogger(logger)
    fh = logging.FileHandler(path)

    fh.setLevel(getattr(logging, level.upper()))

    formatter = logging.Formatter(fmt, "%Y-%m-%d %H:%M:%S")
    fh.setFormatter(formatter)

    log.addHandler(fh)



"""
# Hamiltonian/Hybrid Monte Carlo (HMC)
"""

from dataclasses import dataclass, field
from functools import partial
from typing import Callable, ClassVar, Sequence

import jax.numpy as jnp
from blackjax.adaptation.step_size import find_reasonable_step_size
from blackjax.mcmc import hmc
from jax.flatten_util import ravel_pytree

from .da import da_finalize, da_init, da_step
from .epoch import EpochState
from .kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    ModelMixin,
    TransitionMixin,
    TransitionOutcome,
    TuningMixin,
    TuningOutcome,
    WarmupOutcome,
)
from .mm import tune_inv_mm_diag, tune_inv_mm_full
from .pytree import register_dataclass_as_pytree
from .types import Array, KeyArray, ModelState, Position


@register_dataclass_as_pytree
@dataclass
class HMCKernelState:
    """
    A dataclass for the state of a `HMCKernel`, implementing the
    `liesel.goose.da.DAKernelState` protocol.
    """

    step_size: float
    inverse_mass_matrix: Array
    error_sum: float = field(init=False)
    log_avg_step_size: float = field(init=False)
    mu: float = field(init=False)

    def __post_init__(self):
        da_init(self)


@register_dataclass_as_pytree
@dataclass
class HMCTransitionInfo(DefaultTransitionInfo):
    divergent: bool
    """
    Whether the difference in energy between the original and the new state exceeded
    the divergence threshold of 1000.
    """


def _goose_info(hmc_info: hmc.HMCInfo) -> HMCTransitionInfo:
    error_code = 1 * hmc_info.is_divergent
    acceptance_prob = hmc_info.acceptance_probability
    position_moved = hmc_info.is_accepted

    return HMCTransitionInfo(
        error_code,
        acceptance_prob,
        position_moved,
        hmc_info.is_divergent,
    )


HMCTuningInfo = DefaultTuningInfo


class HMCKernel(
    ModelMixin,
    TransitionMixin[HMCKernelState, HMCTransitionInfo],
    TuningMixin[HMCKernelState, HMCTuningInfo],
):
    """
    A HMC kernel with dual averaging and an inverse mass matrix tuner,
    implementing the `liesel.goose.types.Kernel` protocol.
    """

    error_book: ClassVar[dict[int, str]] = {0: "no errors", 1: "divergent transition"}
    needs_history: ClassVar[bool] = True
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
        initial_step_size: float | None = None,
        initial_inverse_mass_matrix: Array | None = None,
        num_integration_steps: int = 10,
        da_target_accept: float = 0.8,
        da_gamma: float = 0.05,
        da_kappa: float = 0.75,
        da_t0: int = 10,
        mm_diag: bool = True,
    ):
        self.position_keys = tuple(position_keys)
        self._model = None

        self.initial_step_size = initial_step_size
        self.initial_inverse_mass_matrix = initial_inverse_mass_matrix
        self.num_integration_steps = num_integration_steps

        self.da_target_accept = da_target_accept
        self.da_gamma = da_gamma
        self.da_kappa = da_kappa
        self.da_t0 = da_t0

        self.mm_diag = mm_diag

    def _blackjax_state(self, model_state: ModelState) -> hmc.HMCState:
        return hmc.init(self.position(model_state), self.log_prob_fn(model_state))

    def _blackjax_kernel(self) -> Callable:
        return hmc.kernel()

    def init_state(self, prng_key, model_state):
        """
        Initializes the kernel state with an identity inverse mass matrix
        and a reasonable step size (unless explicit arguments were provided
        by the user).
        """

        if self.initial_inverse_mass_matrix is None:
            flat_position, _ = ravel_pytree(self.position(model_state))

            if self.mm_diag:
                inverse_mass_matrix = jnp.ones_like(flat_position)
            else:
                inverse_mass_matrix = jnp.eye(flat_position.size)
        else:
            inverse_mass_matrix = self.initial_inverse_mass_matrix

        if self.initial_step_size is None:
            blackjax_kernel = self._blackjax_kernel()
            blackjax_state = self._blackjax_state(model_state)
            log_prob_fn = self.log_prob_fn(model_state)

            def kernel_generator(step_size: float) -> Callable:
                return partial(
                    blackjax_kernel,
                    logprob_fn=log_prob_fn,
                    step_size=step_size,
                    inverse_mass_matrix=inverse_mass_matrix,
                    num_integration_steps=self.num_integration_steps,
                )

            step_size = find_reasonable_step_size(
                prng_key,
                kernel_generator,
                blackjax_state,
                initial_step_size=0.001,
                target_accept=self.da_target_accept,
            )
        else:
            step_size = self.initial_step_size

        return HMCKernelState(step_size, inverse_mass_matrix)

    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[HMCKernelState, HMCTransitionInfo]:
        """
        Performs an MCMC transition *without* dual averaging.
        """

        blackjax_kernel = self._blackjax_kernel()
        blackjax_state = self._blackjax_state(model_state)
        log_prob_fn = self.log_prob_fn(model_state)

        blackjax_state, blackjax_info = blackjax_kernel(
            prng_key,
            blackjax_state,
            log_prob_fn,
            kernel_state.step_size,
            kernel_state.inverse_mass_matrix,
            self.num_integration_steps,
        )

        info = _goose_info(blackjax_info)
        model_state = self.model.update_state(blackjax_state.position, model_state)
        return TransitionOutcome(info, kernel_state, model_state)

    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[HMCKernelState, HMCTransitionInfo]:
        """
        Performs an MCMC transition *with* dual averaging.
        """

        outcome = self._standard_transition(prng_key, kernel_state, model_state, epoch)

        da_step(
            outcome.kernel_state,
            outcome.info.acceptance_prob,
            epoch.time_in_epoch,
            self.da_target_accept,
            self.da_gamma,
            self.da_kappa,
            self.da_t0,
        )

        return outcome

    def _tune_fast(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[HMCKernelState, HMCTuningInfo]:
        """
        Currently does nothing.
        """

        info = HMCTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def _tune_slow(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[HMCKernelState, HMCTuningInfo]:
        """
        Tunes the inverse mass vector or matrix using the samples from the last epoch.
        """

        if history is not None:
            history = Position({k: history[k] for k in self.position_keys})

            if self.mm_diag:
                new_inv_mm = tune_inv_mm_diag(history)
                trace_fn = jnp.sum
            else:
                new_inv_mm = tune_inv_mm_full(history)
                trace_fn = jnp.trace

            old_inv_mm = kernel_state.inverse_mass_matrix
            adjustment = jnp.sqrt(trace_fn(old_inv_mm) / trace_fn(new_inv_mm))
            kernel_state.step_size = adjustment * kernel_state.step_size

            kernel_state.inverse_mass_matrix = new_inv_mm

        return self._tune_fast(prng_key, kernel_state, model_state, epoch, history)

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> HMCKernelState:
        """
        Resets the state of the dual averaging algorithm.
        """

        da_init(kernel_state)
        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> HMCKernelState:
        """
        Sets the step size as found by the dual averaging algorithm.
        """

        da_finalize(kernel_state)
        return kernel_state

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: HMCKernelState,
        model_state: ModelState,
        tuning_history: HMCTuningInfo | None,
    ) -> WarmupOutcome[HMCKernelState]:
        """
        Currently does nothing.
        """

        return WarmupOutcome(error_code=0, kernel_state=kernel_state)



"""
# Bivariate copulas
"""

import numpy as np
import tensorflow_probability.substrates.numpy.bijectors as tfb
import tensorflow_probability.substrates.numpy.distributions as tfd
from tensorflow_probability.substrates.numpy.internal import parameter_properties

from ..bijectors import AlgebraicSigmoid

__docformat__ = "google"


class GaussianCopula(tfd.TransformedDistribution):
    """A bivariate Gaussian copula."""

    def __init__(
        self,
        dependence=None,
        validate_args=False,
        allow_nan_stats=True,
        name="GaussianCopula",
    ):
        """
        Constructs the distribution.

        Args:
            dependence: The correlation parameter.
        """

        parameters = dict(locals())

        batch_shape = np.shape(dependence)
        loc = np.zeros(batch_shape + (2,))

        if dependence is None:
            scale_tril = None
        else:
            if validate_args:
                assert np.all(dependence >= 0.0)
                assert np.all(dependence <= 1.0)

            tril11 = np.broadcast_to(1.0, batch_shape)
            tril12 = np.broadcast_to(0.0, batch_shape)
            tril22 = np.sqrt(1.0 - dependence**2)

            tril1 = np.stack([tril11, tril12], axis=-1)
            tril2 = np.stack([dependence, tril22], axis=-1)
            scale_tril = np.stack([tril1, tril2], axis=-2)

        distribution = tfd.MultivariateNormalTriL(
            loc=loc,
            scale_tril=scale_tril,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
        )

        bijector = tfb.NormalCDF(validate_args=validate_args)

        super().__init__(
            distribution=distribution,
            bijector=bijector,
            validate_args=validate_args,
            name=name,
        )

        self._parameters = parameters

    @classmethod
    def _parameter_properties(cls, dtype, num_classes=None):
        return {
            "dependence": parameter_properties.ParameterProperties(
                shape_fn=lambda sample_shape: sample_shape[:-1],
                default_constraining_bijector_fn=lambda: AlgebraicSigmoid(),
            )
        }



"""
# Gibbs sampler
"""

from typing import Callable, ClassVar, Sequence

from .epoch import EpochState
from .kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    ModelMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from .types import Kernel, KernelState, KeyArray, ModelState, Position, TuningInfo

GibbsKernelState = KernelState
GibbsTransitionInfo = DefaultTransitionInfo
GibbsTuningInfo = DefaultTuningInfo


class GibbsKernel(
    ModelMixin, Kernel[GibbsKernelState, GibbsTransitionInfo, GibbsTuningInfo]
):
    """
    A Gibbs kernel implementing the `liesel.goose.types.Kernel` protocol.
    """

    error_book: ClassVar[dict[int, str]] = {0: "no errors"}
    needs_history: ClassVar[bool] = False
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
        transition_fn: Callable[[KeyArray, ModelState], Position],
    ):
        self._model = None
        self.position_keys = tuple(position_keys)
        self._transition_fn = transition_fn

    def init_state(self, prng_key, model_state):
        """
        Initializes an (empty) kernel state.
        """

        return {}

    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: KernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[KernelState, GibbsTransitionInfo]:
        """
        Performs an MCMC transition.
        """

        info = GibbsTransitionInfo(
            error_code=0,
            acceptance_prob=1.0,
            position_moved=1,
        )

        position = self._transition_fn(prng_key, model_state)
        model_state = self.model.update_state(position, model_state)
        return TransitionOutcome(info, kernel_state, model_state)

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: KernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None = None,
    ) -> TuningOutcome[KernelState, GibbsTuningInfo]:
        """
        Currently does nothing.
        """

        info = GibbsTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: KernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> KernelState:
        """
        Currently does nothing.
        """

        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: KernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> KernelState:
        """
        Currently does nothing.
        """

        return kernel_state

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: KernelState,
        model_state: ModelState,
        tuning_history: TuningInfo | None,
    ) -> WarmupOutcome[KernelState]:
        """
        Currently does nothing.
        """

        return WarmupOutcome(error_code=0, kernel_state=kernel_state)



"""
# Posterior statistics and diagnostics
"""

import typing
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import arviz as az
import jax.numpy as jnp
import jaxlib.xla_extension
import numpy as np
import pandas as pd
import xarray

from liesel.goose.engine import SamplingResult
from liesel.goose.pytree import slice_leaves, stack_leaves
from liesel.goose.types import Position


def raise_chain_indices_error(
    chain_indices: Sequence[int], num_original_chains: int
) -> None:
    """Display informative error message with valid `chain_indices` inputs."""
    if any(
        chain_index not in range(num_original_chains) for chain_index in chain_indices
    ):
        raise ValueError(
            f"All chain indices must be between 0 and {num_original_chains-1} "
            "(bounds inclusive)."
        )


def validate_chain_indices(
    chain_indices: int | Sequence[int] | None,
    num_original_chains: int,
) -> Sequence[int]:
    """Convert `int` or `None` input of `chain_indices` to sequence of integers."""
    if chain_indices is None:
        return list(range(num_original_chains))

    if isinstance(chain_indices, int):
        chain_indices = [chain_indices]

    raise_chain_indices_error(chain_indices, num_original_chains)
    return chain_indices


def numpy_to_arviz(subparam_chains: np.ndarray) -> xarray.DataArray:
    """Convert data structure of `arviz` package to numpy array."""
    arviz_data = az.convert_to_inference_data(subparam_chains)
    return arviz_data["posterior"]["x"]


def add_rhat(arviz_array: xarray.DataArray, round_digits: int) -> np.ndarray:
    """Compute a single Rhat value for multiple chains of the same subparameter."""
    return np.round(az.rhat(arviz_array)["x"].values, decimals=round_digits)


def combine_chains(subparam_chains: np.ndarray) -> tuple[np.ndarray, xarray.DataArray]:
    """Concatenate all separate chains to a single chain for multi-chain sampling."""
    subparam_chains = subparam_chains.reshape(1, -1)
    arviz_array = numpy_to_arviz(subparam_chains)

    return subparam_chains, arviz_array


def add_num_effective(arviz_array: xarray.DataArray, round_digits: int) -> np.ndarray:
    """
    Compute the effective sample size of one specific subparameter (e.g. `beta_0`).
    """
    return np.array(
        [az.ess(arviz_array[[i]])["x"].values for i in range(arviz_array.shape[0])]
    ).round(decimals=round_digits)


def compute_quantiles(
    subparam_chains: np.ndarray, quantiles: Sequence[float], round_digits: int
) -> np.ndarray:
    """
    Compute posterior quantiles of one specific subparameter (e.g. `beta_0`).
    """
    # i'th row contains i'th input quantile for all chains
    # j'th column contains all input quantiles for j'th chain
    return np.quantile(subparam_chains, q=quantiles, axis=1).round(round_digits)


def add_quantiles(
    subparam_chains: np.ndarray,
    subparam_stats: dict[str, int | float | np.ndarray],
    quantiles: Sequence[float],
    round_digits: int,
) -> dict[str, int | float | np.ndarray]:
    """
    Add posterior quantiles to dictionary which collects all summary statistics and
    diagnostics for one specific subparameter (e.g. `beta_0`).
    """
    # number of rows equals length of quantiles input
    quantiles_array = compute_quantiles(subparam_chains, quantiles, round_digits)

    for ind, q in enumerate(quantiles):
        subparam_stats[f"q_{100 * q:.5g}"] = quantiles_array[ind]

    return subparam_stats


def compute_single_hdi(
    arviz_subarray: xarray.DataArray, hdi_prob: float, round_digits: int
) -> np.ndarray:
    """Compute Highest-Density Interval for a single chain."""
    # compute separate hdi for each chain to stay consistent with credible intervals
    # from `add_quantile()`
    return np.round(
        az.hdi(arviz_subarray, hdi_prob=hdi_prob)["x"].values, decimals=round_digits
    )


def compute_hdi(
    arviz_array: xarray.DataArray, hdi_prob: float, round_digits: int
) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute Highest-Density Intervals for all chains. Returns two numpy arrays of lower
    and upper bounds of Highest-Density Intervals.
    """
    # numpy array where i'th row contains lower and upper hdi bound of i'th chain
    hdi = np.vstack(
        [
            # subset with list to keep first dimension, arviz needs 2d arrays
            compute_single_hdi(arviz_array[[i]], hdi_prob, round_digits)
            for i in range(arviz_array.shape[0])
        ]
    )

    return hdi[:, 0], hdi[:, 1]


def add_hdi(
    arviz_array: xarray.DataArray,
    subparam_stats: dict[str, int | float | np.ndarray],
    hdi_prob: float,
    round_digits: int,
) -> dict[str, int | float | np.ndarray]:
    """
    Add Highest Density Intervals to dictionary which collects all summary statistics
    and diagnostics for one specific subparameter (e.g. `beta_0`).
    """
    hdi_lower, hdi_upper = compute_hdi(arviz_array, hdi_prob, round_digits)

    subparam_stats[f"hdi_{100 * hdi_prob:.5g}_low"] = hdi_lower
    subparam_stats[f"hdi_{100 * hdi_prob:.5g}_high"] = hdi_upper

    return subparam_stats


def subparam_stats_to_df(
    subparam_stats: dict[str, int | float | np.ndarray], per_chain: bool
) -> pd.DataFrame:
    """
    Convert dictionary which collects all summary statistics and diagnostics for one
    specific subparameter (e.g. `beta_0`) to a `pandas` data frame.
    """
    if not per_chain:
        # Must be dropped BEFORE conversion to data frame due to unequal lengths
        del subparam_stats["chain_index"]

    return pd.DataFrame(subparam_stats).drop(columns="num_chains")


def get_subparam_stats(
    subparam_chains: np.ndarray,
    per_chain: bool,
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
    param_index: int,
    as_dataframe: bool,
) -> dict[str, int | float | np.ndarray] | pd.DataFrame:
    """
    Collect all summary statistics and diagnostics for one specific subparameter (e.g.
    `beta_0`) into one dictionary or `pandas` data frame.
    """
    num_original_chains = subparam_chains.shape[0]
    chain_indices = validate_chain_indices(chain_indices, num_original_chains)

    subparam_chains = subparam_chains[chain_indices]
    num_filtered_chains = subparam_chains.shape[0]
    num_samples: int | list[int] = [subparam_chains.shape[1]] * num_filtered_chains

    arviz_array = numpy_to_arviz(subparam_chains)
    rhat = add_rhat(arviz_array, round_digits)

    if not per_chain:
        subparam_chains, arviz_array = combine_chains(subparam_chains)
        num_samples = sum(num_samples)  # type: ignore

    num_effective = add_num_effective(arviz_array, round_digits)
    posterior_means = subparam_chains.mean(axis=1).round(round_digits)
    posterior_sds = subparam_chains.std(axis=1).round(round_digits)

    subparam_stats = {
        "param_index": param_index,
        "num_chains": num_filtered_chains,
        "chain_index": chain_indices,
        "num_samples": num_samples,
        "num_effective": num_effective,
        "mean": posterior_means,
        "sd": posterior_sds,
        "rhat": rhat,
    }

    subparam_stats = add_quantiles(
        subparam_chains, subparam_stats, quantiles, round_digits
    )
    subparam_stats = add_hdi(arviz_array, subparam_stats, hdi_prob, round_digits)

    if not as_dataframe:
        return subparam_stats

    return subparam_stats_to_df(subparam_stats, per_chain)


def raise_dimension_error(param: str, num_dim: int) -> None:
    """Check for correct array dimensions of posterior samples."""
    if num_dim not in (2, 3):
        raise ValueError(
            f"Array of posterior samples for {param} has the wrong number of"
            f"dimensions.\nExpected 2 or 3, got {num_dim}."
        )


def adjust_dimensions(param_chains: np.ndarray, num_dim: int) -> np.ndarray:
    """
    Make shape of posterior samples for one dimensional parameters (e.g. `log_sigma`)
    consistent with multi-dimensional parameters.
    """
    if num_dim == 2:
        param_chains = np.expand_dims(param_chains, axis=-1)
    return param_chains


def raise_param_indices_error(
    param_indices: Sequence[int], num_original_subparams: int, param: str
) -> None:
    """
    Display informative error message with valid `param_indices` inputs for this
    specific `param`.
    """
    if any(
        param_index not in range(num_original_subparams)
        for param_index in param_indices
    ):
        raise ValueError(
            f"All param indices for {param} must be between "
            f"0 and {num_original_subparams-1} (bounds inclusive)."
        )


def validate_param_indices(
    param_indices: int | Sequence[int] | None, num_original_subparams: int, param: str
) -> Sequence[int]:
    """Convert `int` or `None` input of `param_indices` to sequence of integers."""
    if param_indices is None:
        return list(range(num_original_subparams))

    if isinstance(param_indices, int):
        param_indices = [param_indices]

    raise_param_indices_error(param_indices, num_original_subparams, param)

    return param_indices


def collect_subparam_dicts(
    param_chains: np.ndarray,
    per_chain: bool,
    param_indices: Sequence[int],
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
) -> list[dict]:
    """
    Combine dictionaries with summary statistics and diagnostics within one parameter
    vector (e.g. `beta_0` and `beta_1`) into a list.
    """
    return [
        get_subparam_stats(
            param_chains[..., i],
            per_chain,
            chain_indices,
            quantiles,
            hdi_prob,
            round_digits,
            param_index=i,
            as_dataframe=False,
        )
        for i in param_indices
    ]


def move_col_first(df: pd.DataFrame, colname: str) -> pd.DataFrame:
    """Move last column of a `pandas` data frame to the first column."""
    return df[[colname] + [col for col in df.columns if col != colname]]


def collect_subparam_dfs(
    param_chains: np.ndarray,
    per_chain: bool,
    param_indices: Sequence[int],
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
) -> pd.DataFrame:
    """
    Combine data frames with summary statistics and diagnostics within one parameter
    vector (e.g. `beta_0` and `beta_1`) into a single `pandas` data frame.
    """
    param_df = pd.concat(
        [
            get_subparam_stats(
                param_chains[..., i],
                per_chain,
                chain_indices,
                quantiles,
                hdi_prob,
                round_digits,
                param_index=i,
                as_dataframe=True,
            )
            for i in param_indices
        ]
    ).reset_index(drop=True)

    return move_col_first(param_df, "param_index")


def get_param_stats(
    param: str,
    posterior_samples: dict[str, jnp.DeviceArray],
    per_chain: bool,
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
    as_dataframe: bool,
) -> list[dict] | pd.DataFrame:
    """
    Collect all summary statistics and diagnostics for one specific parameter vector
    (e.g. `beta`) into one dictionary or `pandas` data frame.
    """
    # arviz package requires numpy arrays instead of jax numpy arrays
    param_chains = np.array(posterior_samples[param])
    num_dim = param_chains.ndim

    # check that each param_chain has either two (scalar) or three (vector) dimension
    raise_dimension_error(param, num_dim)

    # add last dimension if original parameter is a scalar instead of a vector
    param_chains = adjust_dimensions(param_chains, num_dim)

    num_original_subparams = param_chains.shape[-1]
    param_indices = validate_param_indices(param_indices, num_original_subparams, param)

    if not as_dataframe:
        return collect_subparam_dicts(
            param_chains,
            per_chain,
            param_indices,
            chain_indices,
            quantiles,
            hdi_prob,
            round_digits,
        )

    return collect_subparam_dfs(
        param_chains,
        per_chain,
        param_indices,
        chain_indices,
        quantiles,
        hdi_prob,
        round_digits,
    )


def validate_params(
    posterior_samples: dict[str, jnp.DeviceArray], params: str | list[str] | None
) -> list[str]:
    """Convert `str` or `None` input of `params` to sequence of strings."""
    posterior_keys = list(posterior_samples.keys())
    if params is None:
        return posterior_keys

    if isinstance(params, str):
        params = [params]

    if any(param not in posterior_keys for param in params):
        raise KeyError(f"All params must be in {posterior_keys}.")

    return params


def collect_param_dicts(
    posterior_samples: dict[str, jnp.DeviceArray],
    per_chain: bool,
    params: list[str],
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
) -> dict[str, list[dict]]:
    """
    Combine dictionaries with summary statistics and diagnostics of all model parameters
    into a single dictionary.
    """
    return {
        param: get_param_stats(
            param,
            posterior_samples,
            per_chain,
            param_indices,
            chain_indices,
            quantiles,
            hdi_prob,
            round_digits,
            as_dataframe=False,
        )
        for param in params
    }


def collect_param_dfs(
    posterior_samples: dict[str, jnp.DeviceArray],
    per_chain: bool,
    params: list[str],
    param_indices: int | Sequence[int] | None,
    chain_indices: int | Sequence[int] | None,
    quantiles: Sequence[float],
    hdi_prob: float,
    round_digits: int,
) -> pd.DataFrame:
    """
    Combine data frames with summary statistics and diagnostics of all model parameters
    into a single `pandas` data frame.
    """
    param_dfs: list[pd.DataFrame] = []
    for param in params:
        param_df: pd.DataFrame = get_param_stats(
            param,
            posterior_samples,
            per_chain,
            param_indices,
            chain_indices,
            quantiles,
            hdi_prob,
            round_digits,
            as_dataframe=True,
        )
        param_df = param_df.assign(param=param).set_index("param")

        param_df.index.name = None
        param_dfs.append(param_df)

    return pd.concat(param_dfs)


def summary(
    results: SamplingResult,
    per_chain: bool = True,
    params: str | list[str] | None = None,
    param_indices: int | Sequence[int] | None = None,
    chain_indices: int | Sequence[int] | None = None,
    quantiles: Sequence[float] = (0.05, 0.5, 0.95),
    hdi_prob: float = 0.9,
    round_digits: int = 3,
    as_dataframe: bool = True,
) -> dict[str, list[dict]] | pd.DataFrame:
    """
    Compute summary statistics and diagnostic measures of posterior samples.

    ## Parameters

    - `results`:
      Result object of the sampling process. Must have a method
      `get_posterior_samples()` which extracts all samples from the posterior
      distribution.

    - `per_chain`:
      If `True`, all statistics and diagnostics (except the Rhat value) are computed for
      each chain separately. If `False`, one metric is computed for each subparameter
      after concatenating all chains.

    - `params`:
      Names of the model parameters that are contained in the summary output. Must
      coincide with the dictionary keys of the `Position` with the posterior samples.
      If `None`, all parameters are included.

    - `param_indices`:
      Indices of each model parameter that are contained in the summary output. Selects
      e.g. `beta_0` out of a `beta` parameter vector.A single index can be specified as
      an integer or a sequence containing one integer. If `None`, all subparameters are
      included.

    - `chain_indices`:
      Indices of chains for each model subparameter that are contained in the summary
      output. Selects e.g. chain 0 and chain 2 out of multiple chains. A single index
      can be specified as an integer or a sequence containing one integer. If `None`,
      all chains are included.

    - `quantiles`:
      Quantiles of the posterior distribution that are contained in the summary output.

    - `hdi_prob`:
      Coverage level of the Highest Density Interval of the posterior distribution.
      Summary output contains lower and upper bound of this interval.

    - `round_digits`:
      Number of decimals for each float value within the summary output.

    - `as_dataframe`:
      If `True`, all statistics and diagnostics are embedded into a `pandas` data frame.
      If `False`, a dictionary with the same keys as `Position` with the posterior
      samples is returned.

    ## Returns

    Dictionary if `as_dataframe` is False, `pandas.DataFrame` if `as_dataframe` is
    True.
    """
    # NOTE: Docstring Duplications
    # The entries `results`, `params`, `param_indices` and `chain_indices` are shared
    # with all user plotting functions.
    posterior_samples = results.get_posterior_samples()
    params = validate_params(posterior_samples, params)

    if not as_dataframe:
        return collect_param_dicts(
            posterior_samples,
            per_chain,
            params,
            param_indices,
            chain_indices,
            quantiles,
            hdi_prob,
            round_digits,
        )

    return collect_param_dfs(
        posterior_samples,
        per_chain,
        params,
        param_indices,
        chain_indices,
        quantiles,
        hdi_prob,
        round_digits,
    )


@dataclass
class Summary:
    """
    A summary object.

    Allows easy programatic access via quantities[quantity_name][var_name]. The
    array has the a similar shape as the parameter with var_name. However, if
    `per_chain` is `True` the first dimension refers to the chain index.
    Additionally, for "hdi" and "quantile" the second dimension refers to the
    quantile.

    The summary object can be turned into a `pd.DataFrame` using the function
    `to_dataframe`.

    Experimental.
    """

    quantities: dict[str, dict[str, np.ndarray]]
    config: dict
    sample_info: dict

    def to_dataframe(self) -> pd.DataFrame:
        """Turns Summary object into a DataFrame object."""

        # don't change the original data
        quants = self.quantities.copy()

        # make new entries for the quantiles
        for i, q in enumerate(self.config["quantiles"]):
            quants[f"q_{q}"] = {k: v[i, ...] for k, v in quants["quantile"].items()}
        quants["hdi_low"] = {k: v[0, ...] for k, v in quants["hdi"].items()}
        quants["hdi_high"] = {k: v[1, ...] for k, v in quants["hdi"].items()}

        # remove the old entries
        del quants["quantile"]
        del quants["hdi"]

        # create one row per entry
        df_dict = {}
        for var in quants["mean"].keys():
            it = np.nditer(quants["mean"][var], flags=["multi_index"])
            for _ in it:
                var_fqn = (
                    var if len(it.multi_index) == 0 else f"{var}{list(it.multi_index)}"
                )
                quant_per_elem: dict[str, Any] = {}
                quant_per_elem["variable"] = var
                if self.config["chains_merged"]:
                    quant_per_elem["var_index"] = it.multi_index
                    quant_per_elem["sample_size"] = (
                        self.sample_info["sample_size_per_chain"]
                        * self.sample_info["num_chains"]
                    )
                else:
                    quant_per_elem["chain_index"] = it.multi_index[0]
                    quant_per_elem["var_index"] = it.multi_index[1:]
                    quant_per_elem["sample_size"] = self.sample_info[
                        "sample_size_per_chain"
                    ]

                for quant_name, quant_dict in quants.items():
                    quant_per_elem[quant_name] = quant_dict[var][it.multi_index]

                # convert DeviceArrays (scalar) to floats so that
                # pandas treats them correctly

                for key, val in quant_per_elem.items():
                    if type(val) == jaxlib.xla_extension.DeviceArray:
                        # make mypy happy
                        val = typing.cast(jnp.ndarray, val)
                        # value should be a scalar
                        assert val.shape == ()

                        # convert to float32
                        val = np.atleast_1d(np.asarray(val))[0]

                        quant_per_elem[key] = val

                df_dict[var_fqn] = quant_per_elem

        # convert to dataframe and use varname as index
        df = pd.DataFrame.from_dict(df_dict, orient="index")
        df = df.reset_index()
        df = df.rename(columns={"index": "var_fqn"})
        df = df.set_index("variable")

        return df

    def _repr_html_(self):
        df = self.to_dataframe()
        del df["var_fqn"]
        return df.to_html(index_names=False)

    def __str__(self):
        return self.to_dataframe().__str__()

    @staticmethod
    def from_result(
        result: SamplingResult,
        additional_chain: Position | None = None,
        quantiles: Sequence[float] = (0.05, 0.5, 0.95),
        hdi_prob: float = 0.9,
        selected: list[str] | None = None,
        deselected: list[str] | None = None,
        per_chain=False,
    ) -> "Summary":
        """
        Creates a `Summary` object from a result object.

        An optional `addition_chain` can be supplied to add more parameters to
        the summary output. `additional_chain` must be a position chain which
        maches chain and time dimention of the posterior chain as returned by
        `result.get_posterior_sampels()`.

        The arguments `selected` and `deselected` allow to get a summary only
        from subset of the position keys.

        When using `per_chain` the summary is calculated on a per chain basis.
        Certain measures like rhat are not availavle if `per_chain` is true.

        Experimental.
        """

        posterior_chain = result.get_posterior_samples()
        if additional_chain:
            for k, v in additional_chain.items():
                posterior_chain[k] = v

        if selected:
            posterior_chain = Position(
                {
                    key: value
                    for key, value in posterior_chain.items()
                    if key in selected
                }
            )

        if deselected is not None:
            for key in deselected:
                del posterior_chain[key]

        # get some general infos on the sampling
        param_chain = next(iter(posterior_chain.values()))
        sample_info = {
            "num_chains": param_chain.shape[0],
            "sample_size_per_chain": param_chain.shape[1],
        }

        # convert everything to numpy array
        for key in posterior_chain:
            posterior_chain[key] = np.asarray(posterior_chain[key])

        # calculate quantiles either per chain and merge the results or all at once
        single_chain_summaries = []
        if per_chain:
            for chain_idx in range(sample_info["num_chains"]):
                single_chain = slice_leaves(
                    posterior_chain, jnp.s_[None, chain_idx, ...]
                )
                single_chain_summaries.append(
                    _create_quantity_dict(single_chain, quantiles, hdi_prob)
                )
            quantities = stack_leaves(single_chain_summaries, axis=0)
        else:
            quantities = _create_quantity_dict(posterior_chain, quantiles, hdi_prob)

        config = {
            "quantiles": quantiles,
            "hdi_prob": hdi_prob,
            "chains_merged": not per_chain,
        }

        return Summary(quantities=quantities, config=config, sample_info=sample_info)


def _create_quantity_dict(
    chain: Position, quantiles: Sequence[float], hdi_prob: float
) -> dict[str, dict[str, np.ndarray]]:
    azchain = az.convert_to_inference_data(chain).posterior

    # calculate quantities
    mean = azchain.mean(dim=["chain", "draw"])
    var = azchain.var(dim=["chain", "draw"])
    sd = azchain.std(dim=["chain", "draw"])
    quantile = azchain.quantile(q=quantiles, dim=["chain", "draw"])
    hdi = az.hdi(azchain, hdi_prob=hdi_prob)

    ess_bulk = az.ess(azchain, method="bulk")
    ess_tail = az.ess(azchain, method="tail")
    mcse_mean = az.mcse(azchain, method="mean")
    mcse_sd = az.mcse(azchain, method="sd")

    # place quantities in dict
    quantities = {
        "mean": mean,
        "var": var,
        "sd": sd,
        "quantile": quantile,
        "hdi": hdi,
        "rhat": None,
        "ess_bulk": ess_bulk,
        "ess_tail": ess_tail,
        "mcse_mean": mcse_mean,
        "mcse_sd": mcse_sd,
    }

    if azchain.chain.size > 1:
        quantities["rhat"] = az.rhat(azchain)
    else:
        del quantities["rhat"]

    # convert to simple dict[str, np.ndarray]
    for key, val in quantities.items():
        quantities[key] = {k: v.values for k, v in val.data_vars.items()}

    # special treatment for hdi since the function uses the last axis to refer
    # to the quantile
    for k, v in quantities["hdi"].items():
        quantities["hdi"][k] = np.moveaxis(v, -1, 0)

    return quantities



"""
# Model nodes
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from functools import reduce
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    TypeVar,
    cast,
    overload,
)

import jax
import jax.numpy as jnp
import numpy as np

from liesel.option import Option, WeakOption
from liesel.tfp.jax import bijectors as jb
from liesel.tfp.jax import distributions as jd
from liesel.tfp.numpy import bijectors as nb
from liesel.tfp.numpy import distributions as nd

from .goose import make_log_prob_fn
from .types import (
    Array,
    ModelState,
    NodeState,
    Position,
    TFPBijector,
    TFPBijectorClass,
    TFPDistribution,
    TFPDistributionClass,
)

if TYPE_CHECKING:
    from .model import Model


def _input_args(inputs: NodeInputs, num: int | None = None) -> str:
    items = inputs.items()
    args = [f"{node:s}" for key, node in items if isinstance(key, int)]
    kwargs = [f"{key}={node:s}" for key, node in items if isinstance(key, str)]
    return _join_args([_join_args(args, num), _join_args(kwargs, num)])


def _join_args(args: Iterable[str], num: int | None = None) -> str:
    args = [arg for arg in args if arg != ""]

    if num is not None and len(args) > num:
        args = args[0:num] + ["..."]

    return ", ".join(args)


def _num_arg(x: Any, key: str | None = None) -> str:
    if key is not None:
        key = f"{key}="
    else:
        key = ""

    arg = re.sub(r"\((?s:.*)\)", "(...)", repr(x))

    return f"{key}{arg}"


def _opt_arg(x: Any, key: str | None = None) -> str:
    if x is None:
        return ""

    if key is not None:
        key = f"{key}="
    else:
        key = ""

    return f"{key}{repr(x)}"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Node components ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class NodeInputs(tuple):
    """The inputs of a node component as a tuple with optional names."""

    def __new__(cls, *args, **kwargs):
        tup = args + tuple(kwargs.values())
        return super().__new__(cls, tup)

    def __init__(self, *args, **kwargs):
        self._dict = dict(enumerate(args))
        self._dict.update(kwargs)

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return super().__getitem__(key)

        if isinstance(key, str):
            return self._dict[key]

        raise TypeError(f"Indices of {repr(self)} must be integers, slices or strings")

    def __getnewargs__(self):
        return self

    def __repr__(self):
        cls = type(self).__name__
        args = _input_args(self, 1)
        return f"{cls}({args})"

    def dict(self):
        """Returns the node inputs as a dictionary."""

        return self._dict.copy()

    def keys(self):
        """Returns the keys of the node inputs dictionary."""

        return self.dict().keys()

    def values(self):
        """Returns the values of the node inputs dictionary."""

        return self.dict().values()

    def items(self):
        """Returns the items of the node inputs dictionary."""

        return self.dict().items()

    def replace(self, old, new):
        """
        Returns a shallow copy of the node inputs with the `new` node instead of
        the `old` one.
        """

        def fn(node):
            return new if node is old else node

        args = [fn(node) for key, node in self.items() if isinstance(key, int)]
        kwargs = {key: fn(node) for key, node in self.items() if isinstance(key, str)}
        return NodeInputs(*args, **kwargs)


class NodeComponent(ABC):
    """
    An abstract node component with inputs and a NumPy module.

    Superclass of the `NodeDistribution` and the `NodeCalculator`.
    """

    def __init__(self, *args: Node, **kwargs: Node) -> None:
        self._inputs = NodeInputs(*args, **kwargs)
        self._jaxified = False

    @property
    def inputs(self) -> NodeInputs:
        """The inputs of the node component."""

        return self._inputs

    @property
    def jaxified(self) -> bool:
        """Whether JAX NumPy is enabled for the node component."""

        return self._jaxified

    @jaxified.setter
    def jaxified(self, jaxified: bool) -> None:
        if jaxified:
            self.jaxify()
        else:
            self.unjaxify()

    def jaxify(self) -> NodeComponent:
        """Enables JAX NumPy for the node component."""

        self._jaxified = True
        return self

    @property
    def _np(self) -> ModuleType:
        if self.jaxified:
            return jnp

        return np

    def unjaxify(self) -> NodeComponent:
        """Disables JAX NumPy for the node component."""

        self._jaxified = False
        return self

    def __repr__(self) -> str:
        cls = type(self).__name__
        args = _input_args(self.inputs, 1)
        return f"{cls}({args})"


class NodeCalculator(NodeComponent):
    """
    An abstract value calculator of a node.

    Computes the value of a node from its inputs.
    Superclass of the `AdditionCalculator`, the `BijectorCalculator`, etc.
    """

    @abstractmethod
    def value(self) -> Array:
        """Calculates the value of a node from its inputs and returns it."""


class NodeDistribution(NodeComponent):
    """
    A probability distribution of a node.

    Computes the log-probability of a node from its inputs.
    Implemented as a thin wrapper around a TFP distribution.

    ## Parameters

    - `distribution`:
      The name of a TFP distribution as a string, or alternatively, a user-defined
      TFP-compatible distribution class.

      If a class is provided instead of a string, the user needs to make sure it uses
      the right NumPy implementation.

    - `bijector`:
      The name of a TFP bijector as a string, or alternatively, a user-defined
      TFP-compatible bijector class.

      If a class is provided instead of a string, the user needs to make sure it uses
      the right NumPy implementation. Defaults to None.

    - `inputs`:
      The inputs of the distribution.

      The keywords must match the arguments of the TFP distribution.
    """

    def __init__(
        self,
        distribution: str | TFPDistributionClass,
        bijector: str | TFPBijectorClass | None = None,
        **inputs: Node,
    ) -> None:
        super().__init__(**inputs)
        self._distribution = distribution
        self._bijector = bijector

    def distribution(self) -> TFPDistribution:
        """The TFP distribution initialized with the values of the inputs."""

        distribution_module = jd if self.jaxified else nd

        if isinstance(self._distribution, str):
            distribution_cls = getattr(distribution_module, self._distribution)
        else:
            distribution_cls = self._distribution

        kwargs = {kw: node.value for kw, node in self.inputs.items()}
        distribution = distribution_cls(**kwargs)

        if self._bijector is not None:
            bijector_module = jb if self.jaxified else nb

            if isinstance(self._bijector, str):
                bijector_cls = getattr(bijector_module, self._bijector)
            else:
                bijector_cls = self._bijector

            transformed_cls = distribution_module.TransformedDistribution
            distribution = transformed_cls(distribution, bijector_cls())

        return distribution

    def transform(self, bijector: str | TFPBijectorClass) -> NodeDistribution:
        """Transforms the distribution with a TFP bijector."""

        if self._bijector is not None:
            msg = f"Cannot transform {repr(self)}, as it is already transformed"
            raise RuntimeError(msg)

        inputs = self.inputs.dict()

        return NodeDistribution(self._distribution, bijector, **inputs)

    def cdf(self, value: Array, **kwargs) -> Array:
        """
        The cumulative distribution function of the distribution.

        The arguments are passed on to the TFP distribution.
        """

        return self.distribution().cdf(value, **kwargs)

    def log_prob(self, value: Array, **kwargs) -> Array:
        """
        The log-probability (density) function of the distribution.

        The arguments are passed on to the TFP distribution.
        """

        log_probs = self.distribution().log_prob(value, **kwargs)
        return self._np.sum(log_probs)

    def mean(self, **kwargs) -> Array:
        """
        The mean function of the distribution.

        The arguments are passed on to the TFP distribution.
        """

        return self.distribution().mean(**kwargs)

    def sample(self, sample_shape=(), seed=None, **kwargs) -> Array:
        """
        The sampling function of the distribution.

        The arguments are passed on to the TFP distribution.
        """

        return self.distribution().sample(sample_shape, seed, **kwargs)

    def __repr__(self) -> str:
        cls = type(self).__name__

        distribution = repr(self._distribution)
        bijector = _opt_arg(self._bijector)
        inputs = _input_args(self.inputs)

        args = _join_args([distribution, bijector, inputs])

        return f"{cls}({args})"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


GradientFunction = Callable[[Position, ModelState], Position]
TNodeCalculator = TypeVar("TNodeCalculator", bound=NodeCalculator)


class Node(Generic[TNodeCalculator]):
    """
    A node, strong or weak, with or without a probability distribution,
    which can be used to build probabilistic graphical models.

    ## Attributes

    - `outdated`:
      Whether the node is outdated.

      A node is outdated if its value or the value of one of its inputs has changed.
      The value and the log-probability of an outdated node need to be recomputed.

    - `outputs`:
      An output of a node A is a node B which depends on the value of A.
    """

    @overload
    def __init__(
        self: Node[NodeCalculator],
        value: Array,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self: Node[TNodeCalculator],
        value: TNodeCalculator,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        ...

    def __init__(
        self,
        value: Array | TNodeCalculator,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        if isinstance(value, NodeCalculator):
            calculator = cast(TNodeCalculator, value)
            value = 0.0
        else:
            calculator = None

        self._calculator: Option[TNodeCalculator] = Option(calculator)
        self._distribution: Option[NodeDistribution] = Option(distribution)
        self._grad_fn: Option[GradientFunction] = Option(None)
        self._jaxified: bool = False
        self._log_prob: Array = 0.0
        self._model: WeakOption[Model] = WeakOption(None)
        self._name: Option[str] = Option(name)
        self.outdated: bool = True
        self.outputs: set[Node] = set()
        self._value: Array = value

        for input in self.inputs:
            input.outputs.add(self)

        self.update()

    @property
    def calculator(self) -> TNodeCalculator:
        """The calculator of the node."""

        return self._calculator.expect(f"{repr(self)} does not have a calculator")

    @property
    def distribution(self) -> NodeDistribution:
        """The distribution of the node."""

        return self._distribution.expect(f"{repr(self)} does not have a distribution")

    def grad(self) -> Array:
        """
        Returns the gradient of the model log-probability w.r.t. the node value.

        This method is **unlikely to be efficient**. Consider using JAX explicitly
        to JIT-compile more complex functions. Alternatively, you can use Goose.
        """

        if self._grad_fn.is_none():
            self._grad_fn = Option(jax.jit(jax.grad(make_log_prob_fn(self.model))))

        grad_fn = self._grad_fn.unwrap()
        position = Position({self.name: self.value})
        model_state = self.model.state

        grad_dict = grad_fn(position, model_state)
        grad_array = self._np.asarray(grad_dict[self.name])
        return grad_array

    @property
    def has_calculator(self) -> bool:
        """Whether the node has a calculator."""

        return self._calculator.is_some()

    @property
    def has_distribution(self) -> bool:
        """Whether the node has a distribution."""

        return self._distribution.is_some()

    @property
    def has_model(self) -> bool:
        """Whether the node is part of a model."""

        return self._model.is_some()

    @property
    def has_name(self) -> bool:
        """Whether the node has a name."""

        return self._name.is_some()

    def input_value_changed(self) -> Node:
        """
        Informs the node that the value of one of its inputs has changed.

        Flags the node as outdated, and if the node is weak, also flags its outputs
        as outdated.
        """

        self.outdated = True

        if self.weak:
            self.own_value_changed()

        return self

    @property
    def inputs(self) -> set[Node]:
        """All inputs of the node as a set."""

        inputs: set[Node] = set()
        self._calculator.map(lambda x: inputs.update(x.inputs))
        self._distribution.map(lambda x: inputs.update(x.inputs))
        return inputs

    @property
    def jaxified(self) -> bool:
        """Whether JAX NumPy is enabled for the node."""

        return self._jaxified

    @jaxified.setter
    def jaxified(self, jaxified: bool) -> None:
        if jaxified:
            self.jaxify()
        else:
            self.unjaxify()

    def jaxify(self) -> Node:
        """Enables JAX NumPy for the node."""

        self._value = jnp.asarray(self.value)
        self._log_prob = jnp.asarray(self.log_prob)

        self._calculator.map(lambda x: x.jaxify())
        self._distribution.map(lambda x: x.jaxify())

        self._jaxified = True
        return self

    @property
    def log_prob(self) -> Array:
        """The log-probability of the node."""

        return self._log_prob

    @property
    def model(self) -> Model:
        """The model the node is part of."""

        return self._model.expect(f"{repr(self)} is not part of a model")

    @model.setter
    def model(self, model: Model) -> None:
        if self.has_model:
            raise RuntimeError(f"{repr(self)} is already part of a model")

        self._grad_fn = Option(None)
        self._model = WeakOption(model)

    @property
    def name(self) -> str:
        """The name of the node."""

        return self._name.expect(f"{repr(self)} does not have a name")

    @name.setter
    def name(self, name: str) -> None:
        if self.has_model:
            msg = f"Cannot change the name of {repr(self)}, as it is part of a model"
            raise RuntimeError(msg)

        self._name = Option(name)

    @property
    def _np(self) -> ModuleType:
        if self.jaxified:
            return jnp

        return np

    def own_value_changed(self) -> Node:
        """
        Informs the node that its value has changed.

        Flags the node and its outputs as outdated.
        """

        self.outdated = True

        for node in self.outputs:
            node.input_value_changed()

        return self

    def set_value(self, value: Array, update: bool = True) -> Node:
        """
        Sets the value of the node.

        Flags the node and its outputs as outdated, and if requested,
        updates the model.
        """

        if self.weak:
            msg = f"Cannot set the value of {repr(self)}, as it is a weak node"
            raise RuntimeError(msg)

        self._value = value
        self.own_value_changed()

        if update:
            self._model.map(lambda x: x.update())

        return self

    @property
    def state(self) -> NodeState:
        """The value and the log-probability as a `NodeState`."""

        return NodeState(self.value, self.log_prob)

    @state.setter
    def state(self, state: NodeState) -> None:
        self._value, self._log_prob = state

    @property
    def strong(self) -> bool:
        """Whether the node is strong."""

        return self._calculator.is_none()

    def unjaxify(self) -> Node:
        """Disables JAX NumPy for the node."""

        self._value = np.asarray(self.value)
        self._log_prob = np.asarray(self.log_prob)

        self._calculator.map(lambda x: x.unjaxify())
        self._distribution.map(lambda x: x.unjaxify())

        self._jaxified = False
        return self

    def update(self) -> Node:
        """
        Updates the value and the log-probability of the node.

        Assumes that the inputs of the node are up-to-date.
        """

        if self.weak:
            self._value = self.calculator.value()

        if self.has_distribution:
            self._log_prob = self.distribution.log_prob(self.value)

        self.outdated = False
        return self

    def validate(self) -> Node:
        """Checks if the value and the log-probability of the node are finite."""

        assert self._np.all(self._np.isfinite(self.value))
        assert self._np.all(self._np.isfinite(self.log_prob))
        return self

    @property
    def value(self) -> Array:
        """The value of the node."""

        return self._value

    @value.setter
    def value(self, value: Array) -> None:
        self.set_value(value, update=True)

    @property
    def weak(self) -> bool:
        """Whether the node is weak."""

        return self._calculator.is_some()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_grad_fn"] = Option(None)
        state["_model"] = WeakOption(None)
        return state

    def __format__(self, format_spec: str) -> str:
        if format_spec == "s":
            return self.__short_repr__()

        return self.__repr__()

    def __repr__(self) -> str:
        cls = type(self).__name__

        value = _num_arg(self.value)
        calculator = _opt_arg(self._calculator.value)
        distribution = _opt_arg(self._distribution.value)
        name = _opt_arg(self._name.value, "name")

        first_arg = value if self.strong else calculator
        args = _join_args([first_arg, distribution, name])

        if cls != "Node":
            return f"{cls}(Node({args}))"

        return f"Node({args})"

    def __short_repr__(self) -> str:
        cls = type(self).__name__

        if self.has_name:
            return f"{cls}(..., name={repr(self.name)})"

        return f"{cls}(...)"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Node group ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class NodeGroup(dict):
    def __init__(self, name: str | None = None, **nodes: Node) -> None:
        super().__init__(**nodes)
        self._name = Option(name)

    def copy(self) -> NodeGroup:
        """Returns a shallow copy of the node group."""

        return NodeGroup(self._name.value, **self)

    @property
    def has_name(self) -> bool:
        """Whether the node group has a name."""

        return self._name.is_some()

    @property
    def name(self) -> str:
        """The name of the node group."""

        return self._name.expect(f"{repr(self)} does not have a name")

    @name.setter
    def name(self, name: str) -> None:
        self._name = Option(name)

    def __repr__(self) -> str:
        cls = type(self).__name__
        name = _opt_arg(self._name.value)
        nodes = [f"{key}={node:s}" for key, node in self.items()]
        args = _join_args([name, *nodes], 3)
        return f"{cls}({args})"


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Strong nodes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class DesignMatrix(Node):
    """A strong node representing a design matrix."""


class Hyperparameter(Node):
    """A strong node representing a hyperparameter."""


class Response(Node):
    """A strong node representing a response vector or matrix."""


class Parameter(Node):
    """A strong node representing a model parameter."""

    def __init__(
        self,
        value: Array,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        super().__init__(value, distribution, name)

    def initialize_with_mean(self) -> Parameter:
        """Initializes the value of the parameter with its prior mean."""

        def safe_mean() -> Array:
            assert self.has_distribution

            mean = self._np.squeeze(self.distribution.mean())
            mean = self._np.broadcast_to(mean, self._np.shape(self.value))
            assert self._np.all(self._np.isfinite(mean))

            return mean

        try:
            self.value = safe_mean()
        except (AssertionError, NotImplementedError):
            pass

        return self


class RegressionCoef(Parameter):
    """A parameter node representing a vector of regression coefficients."""


class SmoothingParam(Parameter):
    """A parameter node representing a smoothing parameter."""


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak addition nodes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class AdditionCalculator(NodeCalculator):
    """
    Calculates the element-wise sum of its inputs.

    Must have one or more inputs.
    """

    def __init__(self, *inputs: Node) -> None:
        super().__init__(*inputs)

        if len(self.inputs) == 0:
            raise RuntimeError(f"{repr(self)} must have one or more inputs")

    def value(self) -> Array:
        xs = [node.value for node in self.inputs]
        return reduce(self._np.add, xs)


class Addition(Node[AdditionCalculator]):
    """A weak node with an `AdditionCalculator`."""

    def __init__(
        self,
        *inputs: Node,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        calculator = AdditionCalculator(*inputs)
        super().__init__(calculator, distribution, name)


class Predictor(Addition):
    """An addition node representing a regression predictor."""


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak bijector nodes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class BijectorCalculator(NodeCalculator):
    """
    Evaluates the `forward` or the `inverse` method of a TFP bijector at its input.
    """

    def __init__(
        self, bijector: str | TFPBijectorClass, input: Node, inverse: bool = False
    ) -> None:
        super().__init__(input)
        self._bijector = bijector
        self._inverse = inverse

    @property
    def bijector(self) -> TFPBijector:
        module = jb if self.jaxified else nb

        if isinstance(self._bijector, str):
            cls = getattr(module, self._bijector)
        else:
            cls = self._bijector

        return cls()

    def value(self) -> Array:
        x = self.inputs[0].value

        if self._inverse:
            return self.bijector.inverse(x)

        return self.bijector.forward(x)


class Bijector(Node[BijectorCalculator]):
    """A weak node with a `BijectorCalculator`."""

    def __init__(
        self,
        bijector: str | TFPBijectorClass,
        input: Node,
        inverse: bool = False,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        calculator = BijectorCalculator(bijector, input, inverse)
        super().__init__(calculator, distribution, name)


class InverseLink(Bijector):
    """A bijector node representing an inverse link function."""


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak column-stack node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ColumnStackCalculator(NodeCalculator):
    """
    Stacks its inputs column-wise.

    Must have one or more inputs.
    """

    def __init__(self, *inputs: Node) -> None:
        super().__init__(*inputs)

        if len(self.inputs) == 0:
            raise RuntimeError(f"{repr(self)} must have one or more inputs")

    def value(self) -> Array:
        xs = [node.value for node in self.inputs]
        return self._np.column_stack(xs)


class ColumnStack(Node[ColumnStackCalculator]):
    """A weak node with a `ColumnStackCalculator`."""

    def __init__(
        self,
        *inputs: Node,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        calculator = ColumnStackCalculator(*inputs)
        super().__init__(calculator, distribution, name)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak PIT node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class PITCalculator(NodeCalculator):
    """Calculates the probability integral transform of its input."""

    def __init__(self, input, **kwargs):
        super().__init__(_pit_main_input=input, **kwargs)

    def value(self) -> Array:
        input = self.inputs["_pit_main_input"]
        return input.distribution.cdf(input.value)


class PIT(Node[PITCalculator]):
    """A weak node with a `PITCalculator`."""

    def __init__(
        self,
        input: Node,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        kwargs = input.distribution.inputs.dict()
        calculator = PITCalculator(input, **kwargs)
        super().__init__(calculator, distribution, name)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak smooth node ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class SmoothCalculator(NodeCalculator):
    """
    Calculates a smooth `x @ beta`.

    A smooth is the matrix-vector product of a design matrix `x` and a vector of
    regression coefficients `beta`.
    """

    def __init__(self, x, beta):
        super().__init__(x=x, beta=beta)

    def value(self) -> Array:
        x = self.inputs["x"].value
        beta = self.inputs["beta"].value
        return self._np.matmul(x, beta)


class Smooth(Node[SmoothCalculator]):
    """A weak node with a `SmoothCalculator`."""

    def __init__(
        self,
        x: Node,
        beta: Node,
        distribution: NodeDistribution | None = None,
        name: str | None = None,
    ) -> None:
        calculator = SmoothCalculator(x=x, beta=beta)
        super().__init__(calculator, distribution, name)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Parameter transformation ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def transform_parameter(node: Node, bijector: str | TFPBijectorClass) -> NodeGroup:
    """
    Transforms a parameter and its distribution with a TFP bijector.

    Returns a node group of the transformed and the original parameter.
    """

    if node.has_name:
        group_name = f"{node.name}_transformation"
        transformed_name = f"{node.name}_transformed"
        original_name = node.name
    else:
        group_name = "transformation"
        transformed_name = "transformed"
        original_name = "original"

    if isinstance(bijector, str):
        _bijector = getattr(nb, bijector)

    value = _bijector().forward(node.value)
    distribution = node.distribution.transform(bijector)
    transformed = Parameter(value, distribution, transformed_name)
    original = Bijector(bijector, transformed, inverse=True, name=original_name)
    transformed.jaxified = node.jaxified
    original.jaxified = node.jaxified

    return NodeGroup(group_name, transformed=transformed, original=original)



"""
# Type aliases
"""

from typing import Any, NamedTuple

from liesel.goose import Position

__all__ = [
    "Array",
    "ModelState",
    "NodeState",
    "Position",
    "TFPBijector",
    "TFPBijectorClass",
    "TFPDistribution",
    "TFPDistributionClass",
]

Array = Any


class NodeState(NamedTuple):
    """The state of a node as a named tuple."""

    value: Array
    """The value of the node."""

    log_prob: Array
    """The log-probability of the node."""


ModelState = dict[str, NodeState]

TFPBijector = Any
TFPBijectorClass = Any

TFPDistribution = Any
TFPDistributionClass = Any



"""
# Utilities for the IWLS sampler
"""

import jax
import jax.numpy as jnp
import jax.scipy

from .types import Array, KeyArray

triangular_solve = jax.lax.linalg.triangular_solve


def solve(chol_lhs: Array, rhs: Array) -> Array:
    """
    Solves a system of linear equations `lhs @ x = rhs` for x by applying
    forward and backward substitution. Returns x.

    ## Parameters

    - `chol_lhs`: The lower triangular matrix of the Cholesky decomposition
      of the left-hand side of the system.
    - `rhs`: The right-hand side of the system.
    """

    tmp = triangular_solve(chol_lhs, rhs, left_side=True, lower=True)
    return triangular_solve(chol_lhs, tmp, lower=True)


def mvn_log_prob(x: Array, mean: Array, chol_inv_cov: Array) -> Array:
    """
    Returns the log-density of a multivariate normal distribution.

    ## Parameters

    - `x`: The vector of observations.
    - `mean`: The mean vector.
    - `chol_inv_cov`: The lower triangular matrix of the Cholesky decomposition
      of the inverse variance-covariance matrix.
    """

    standardized = (x - mean) @ chol_inv_cov
    log_prob = jnp.sum(jax.scipy.stats.norm.logpdf(standardized))
    adjustment = jnp.sum(jnp.log(jnp.diag(chol_inv_cov)))
    return log_prob + adjustment


def mvn_sample(prng_key: KeyArray, mean: Array, chol_inv_cov: Array) -> Array:
    """
    Returns a random draw from a multivariate normal distribution.

    ## Parameters

    - `prng_key`: The key for JAX' pseudo-random number generator.
    - `mean`: The mean vector.
    - `chol_inv_cov`: The lower triangular matrix of the Cholesky decomposition
      of the inverse variance-covariance matrix.
    """

    standardized = jax.random.normal(prng_key, mean.shape)
    centered = triangular_solve(chol_inv_cov, standardized, lower=True)
    return centered + mean
